#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
# coding: utf-8
from sqlalchemy import (BigInteger, Boolean, Column, DateTime, Float,
                        ForeignKey, Integer, SmallInteger, String, Table, Text,
                        text)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()
metadata = Base.metadata


t_calibration_product_minitile = Table(
    "calibration_product_minitile",
    metadata,
    Column(
        "calibration_product_id",
        ForeignKey("calibration_product.calibration_product_id"),
        primary_key=True,
        nullable=False,
    ),
    Column("minitile_id", ForeignKey("minitile.id"), primary_key=True, nullable=False),
)


class Calibrator(Base):
    __tablename__ = "calibrator"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    sct_id = Column(BigInteger)
    sct_catalog_id = Column(BigInteger)
    type = Column(Integer)

    schedblocks = relationship("Schedblock", secondary="fluxcal_schedblock")
    minitiles = relationship("Minitile", secondary="minitile_calibrator")
    schedblocks1 = relationship("Schedblock", secondary="phasecal_schedblock")
    schedblocks2 = relationship("Schedblock", secondary="polcal_schedblock")


class CalibratorFlux(Base):
    __tablename__ = "calibrator_flux"

    id = Column(Integer, primary_key=True)
    frequency = Column(Float(53))
    obs_date = Column(DateTime)
    flux = Column(Float(53))
    sp_idx = Column(Float(53))
    calibrator_id = Column(ForeignKey("calibrator.id"))
    uv_min = Column(Float(53))
    uv_max = Column(Float(53))

    calibrator = relationship("Calibrator")


class CalibratorPol(Base):
    __tablename__ = "calibrator_pol"

    id = Column(Integer, primary_key=True)
    fraction = Column(Float(53))
    pos_angle = Column(Float(53))
    rm = Column(Float(53))
    ref_freq = Column(Float(53))
    calibrator_id = Column(ForeignKey("calibrator.id"))

    calibrator = relationship("Calibrator")


class ConfigurationFile(Base):
    __tablename__ = "configuration_file"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    content_type = Column(String(255))
    contents = Column(Text)
    jobspec_id = Column(ForeignKey("jobspec.id"), index=True)

    jobspec = relationship("Jobspec")


class ConfigurationFileTemplate(Base):
    __tablename__ = "configuration_file_template"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    content_type = Column(String(255))
    contents = Column(Text)
    doc = Column(Text)
    workflow_id = Column(ForeignKey("workflow.id"))

    workflow = relationship("Workflow")


t_databasechangelog = Table(
    "databasechangelog",
    metadata,
    Column("id", String(255), nullable=False),
    Column("author", String(255), nullable=False),
    Column("filename", String(255), nullable=False),
    Column("dateexecuted", DateTime, nullable=False),
    Column("orderexecuted", Integer, nullable=False),
    Column("exectype", String(10), nullable=False),
    Column("md5sum", String(35)),
    Column("description", String(255)),
    Column("comments", String(255)),
    Column("tag", String(255)),
    Column("liquibase", String(20)),
    Column("contexts", String(255)),
    Column("labels", String(255)),
    Column("deployment_id", String(10)),
)


class Databasechangeloglock(Base):
    __tablename__ = "databasechangeloglock"

    id = Column(Integer, primary_key=True)
    locked = Column(Boolean, nullable=False)
    lockgranted = Column(DateTime)
    lockedby = Column(String(255))


t_fluxcal_schedblock = Table(
    "fluxcal_schedblock",
    metadata,
    Column("schedblock_id", ForeignKey("schedblock.schedblock_id"), primary_key=True, nullable=False),
    Column("calibrator_id", ForeignKey("calibrator.id"), primary_key=True, nullable=False),
)


class InputProductVersion(Base):
    __tablename__ = "input_product_versions"

    id = Column(Integer, primary_key=True)
    product_version_id = Column(ForeignKey("product_version.id"))
    jobspec_id = Column(ForeignKey("jobspec.id"))

    jobspec = relationship("Jobspec")
    product_version = relationship("ProductVersion")
    product_versions = relationship("ProductVersion", secondary="inputs_to_product_versions")


t_inputs_to_product_versions = Table(
    "inputs_to_product_versions",
    metadata,
    Column("intputs_id", ForeignKey("input_product_versions.id"), primary_key=True, nullable=False),
    Column("product_version_id", ForeignKey("product_version.id"), primary_key=True, nullable=False, index=True),
)


class Job(Base):
    __tablename__ = "job"

    id = Column(Integer, primary_key=True)
    start_timestamp = Column(DateTime)
    end_timestamp = Column(DateTime)
    status = Column(Integer)
    json_results = Column(Text)
    qa_analyst = Column(String(255))
    notes = Column(Text)
    jobspec_id = Column(ForeignKey("jobspec.id"), nullable=False, index=True)
    name = Column(String(255))
    archival_job_id = Column(ForeignKey("job.id"), index=True)
    arch_status = Column(Integer)

    archival_job = relationship("Job", remote_side=[id])
    jobspec = relationship("Jobspec")


class JobTaskTemplate(Base):
    __tablename__ = "job_task_template"

    id = Column(Integer, primary_key=True)
    order_num = Column(SmallInteger)
    name = Column(String(255))
    command_template = Column(String(255))
    doc = Column(Text)
    workflow_id = Column(ForeignKey("workflow.id"))

    workflow = relationship("Workflow")


class Jobspec(Base):
    __tablename__ = "jobspec"

    id = Column(Integer, primary_key=True)
    json = Column(Text)
    creation_date = Column(DateTime)
    next_id = Column(ForeignKey("jobspec.id"), index=True)
    queue_id = Column(ForeignKey("queue.id"), index=True)
    name = Column(String(255))
    enqueued = Column(Boolean)
    sdm_id = Column(String(255))

    next = relationship("Jobspec", remote_side=[id])
    queue = relationship("Queue", primaryjoin="Jobspec.queue_id == Queue.id")
    product_versions = relationship("ProductVersion", secondary="jobspec_to_product_version")


t_jobspec_to_product_version = Table(
    "jobspec_to_product_version",
    metadata,
    Column("product_version_id", ForeignKey("product_version.id"), primary_key=True, nullable=False, index=True),
    Column("jobspec_id", ForeignKey("jobspec.id"), primary_key=True, nullable=False),
)


class Jobtask(Base):
    __tablename__ = "jobtask"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    start_timestamp = Column(DateTime)
    end_timestamp = Column(DateTime)
    status = Column(Integer)
    json_results = Column(Text)
    is_final = Column(Boolean)
    job_id = Column(ForeignKey("job.id"), index=True)
    pbs_job_id = Column(String(255))

    job = relationship("Job")


class JsonConfiguration(Base):
    __tablename__ = "json_configuration"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    json = Column(Text)
    product_type_id = Column(ForeignKey("product_type.id"))
    product_version_id = Column(ForeignKey("product_version.id"))
    product_id = Column(ForeignKey("product.id"))

    product = relationship("Product")
    product_type = relationship("ProductType")
    product_version = relationship("ProductVersion")


class Minitile(Base):
    __tablename__ = "minitile"

    id = Column(Integer, primary_key=True, server_default=text("nextval('minitile_id_seq'::regclass)"))
    name = Column(String(255))
    epoch = Column(Integer)
    tier = Column(Integer)
    ra_min = Column(Float(53))
    dec_min = Column(Float(53))
    ra_max = Column(Float(53))
    dec_max = Column(Float(53))
    area = Column(Float(53))
    factor = Column(Float(53))
    first_epoch_half = Column(Boolean)
    custom_1 = Column(String(255))
    custom_2 = Column(String(255))
    custom_3 = Column(String(255))
    custom_4 = Column(String(255))
    custom_5 = Column(String(255))
    definitions = Column(Text)

    schedblocks = relationship("Schedblock", secondary="schedblock_minitile")


t_minitile_calibrator = Table(
    "minitile_calibrator",
    metadata,
    Column("minitile_id", ForeignKey("minitile.id"), primary_key=True, nullable=False),
    Column("calibrator_id", ForeignKey("calibrator.id"), primary_key=True, nullable=False),
)


t_phasecal_schedblock = Table(
    "phasecal_schedblock",
    metadata,
    Column("schedblock_id", ForeignKey("schedblock.schedblock_id"), primary_key=True, nullable=False),
    Column("calibrator_id", ForeignKey("calibrator.id"), primary_key=True, nullable=False),
)


t_polcal_schedblock = Table(
    "polcal_schedblock",
    metadata,
    Column("schedblock_id", ForeignKey("schedblock.schedblock_id"), primary_key=True, nullable=False),
    Column("calibrator_id", ForeignKey("calibrator.id"), primary_key=True, nullable=False),
)


class Prerequisite(Base):
    __tablename__ = "prerequisite"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    ready = Column(Boolean)
    product_id = Column(ForeignKey("product.id"), nullable=False, index=True)
    req_product_id = Column(ForeignKey("product.id"))

    product = relationship("Product", primaryjoin="Prerequisite.product_id == Product.id")
    req_product = relationship("Product", primaryjoin="Prerequisite.req_product_id == Product.id")


class PrerequisiteArchiveid(Base):
    __tablename__ = "prerequisite_archiveids"

    prerequisite_id = Column(ForeignKey("prerequisite.id"), primary_key=True, nullable=False)
    archive_id = Column(String(255), nullable=False)
    name = Column(String(255), primary_key=True, nullable=False)

    prerequisite = relationship("Prerequisite")


class Product(Base):
    __tablename__ = "product"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    category = Column(Integer)
    epoch = Column(Integer)
    notes = Column(Text)
    product_type_id = Column(ForeignKey("product_type.id"), nullable=False)
    completed = Column(Boolean)
    status = Column(Integer, nullable=False, server_default=text("0"))

    product_type = relationship("ProductType")
    tags = relationship("Tag", secondary="product_tag")


class CalibrationProduct(Product):
    __tablename__ = "calibration_product"

    calibration_product_id = Column(ForeignKey("product.id"), primary_key=True)

    minitiles = relationship("Minitile", secondary="calibration_product_minitile")


#
# NOTE:
#   slqacodegen created a much more complete model of the database, but that
#   model included somewhat-circular references among the foreign keys for
#   image_product and it's sub-tables.  This configuration was causing
#   issues using the model within SqlAlchemy for metadata lookups in the
#   service layer (ingestion_metadata.py).
#
#   To postpone this particular part of refactoring the VLASS system, I have
#   deleted image_product and all associated tables, as they weren't required
#   for the current use-case.
#
#   - JLS 08/29/2019
#


class Schedblock(Product):
    __tablename__ = "schedblock"

    schedblock_id = Column(ForeignKey("product.id"), primary_key=True)
    name = Column(String(255))
    opt_id = Column(BigInteger)
    flux_position = Column(Integer)
    lst_start = Column(Float)
    lst_end = Column(Float)
    submitted = Column(Boolean)


t_product_tag = Table(
    "product_tag",
    metadata,
    Column("product_id", ForeignKey("product.id"), primary_key=True, nullable=False),
    Column("tag_id", ForeignKey("tag.id"), primary_key=True, nullable=False),
)


class ProductType(Base):
    __tablename__ = "product_type"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    event_type = Column(String(255))


class ProductVersion(Base):
    __tablename__ = "product_version"

    id = Column(Integer, primary_key=True)
    number = Column(Integer)
    notes = Column(Text)
    product_id = Column(ForeignKey("product.id"))
    status = Column(Integer)

    product = relationship("Product")


class ProductVersionArchiveid(Base):
    __tablename__ = "product_version_archiveids"

    product_version_id = Column(ForeignKey("product_version.id"), primary_key=True, nullable=False)
    archive_id = Column(String(255), nullable=False)
    name = Column(String(255), primary_key=True, nullable=False)

    product_version = relationship("ProductVersion")


class Queue(Base):
    __tablename__ = "queue"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    workflow_id = Column(ForeignKey("workflow.id"))
    head_jobspec_id = Column(ForeignKey("jobspec.id"))
    tail_jobspec_id = Column(ForeignKey("jobspec.id"))
    submit = Column(Boolean)
    project_code = Column(String(255))
    create_job = Column(Boolean)
    auto_qa = Column(Boolean, server_default=text("false"))

    head_jobspec = relationship("Jobspec", primaryjoin="Queue.head_jobspec_id == Jobspec.id")
    tail_jobspec = relationship("Jobspec", primaryjoin="Queue.tail_jobspec_id == Jobspec.id")
    workflow = relationship("Workflow")


t_rawdata = Table(
    "rawdata",
    metadata,
    Column("opt_id", BigInteger),
    Column("sdm_id", String(255)),
    Column("schedblock_id", ForeignKey("schedblock.schedblock_id")),
    Column("rawdata_id", ForeignKey("product.id"), nullable=False),
)


class Scan(Base):
    __tablename__ = "scan"

    id = Column(Integer, primary_key=True)
    start_source_id = Column(ForeignKey("source.id"))
    end_source_id = Column(ForeignKey("source.id"))
    num_phase_centers = Column(Integer)
    minitile_id = Column(ForeignKey("minitile.id"))

    end_source = relationship("Source", primaryjoin="Scan.end_source_id == Source.id")
    minitile = relationship("Minitile")
    start_source = relationship("Source", primaryjoin="Scan.start_source_id == Source.id")


t_schedblock_minitile = Table(
    "schedblock_minitile",
    metadata,
    Column("schedblock_id", ForeignKey("schedblock.schedblock_id"), primary_key=True, nullable=False),
    Column("minitile_id", ForeignKey("minitile.id"), primary_key=True, nullable=False),
)


class Setting(Base):
    __tablename__ = "settings"

    id = Column(String(1), primary_key=True, server_default=text("'X'::bpchar"))
    allow_job_deletion = Column(Boolean, server_default=text("false"))
    allow_product_deletion = Column(Boolean, server_default=text("false"))


class Source(Base):
    __tablename__ = "source"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    sct_catalog_id = Column(BigInteger)
    sct_id = Column(BigInteger)
    ra = Column(Float(53))
    decl = Column(Float(53))


class Tag(Base):
    __tablename__ = "tag"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))


class Workflow(Base):
    __tablename__ = "workflow"

    id = Column(Integer, primary_key=True)
    name = Column(String(255))
    host = Column(String(255))
    port = Column(Integer)
    json = Column(Text)
    enqueue = Column(Boolean)
    product_type_id = Column(ForeignKey("product_type.id"))
    arch_workflow_id = Column(ForeignKey("workflow.id"))

    arch_workflow = relationship("Workflow", remote_side=[id])
    product_type = relationship("ProductType")


class WorkflowProjectCode(Base):
    __tablename__ = "workflow_project_codes"

    project_code_id = Column(ForeignKey("workflow.id"), primary_key=True, nullable=False)
    project_code = Column(String(255), primary_key=True, nullable=False)

    project_code1 = relationship("Workflow")
