#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""  Codifies capability steps, capability request states, execution states,
and execution priorites"""

from __future__ import annotations

from enum import Enum

# pylint: disable=C0103


class CapabilityStepType(Enum):
    """
    Enum that specifies the types of CapabilitySteps that are possible
    """

    PrepareAndRunWorkflow = "prepare-and-run-workflow"
    AwaitQA = "await-qa"
    AwaitWorkflow = "await-workflow"
    Invalid = "invalid"

    @classmethod
    def from_string(cls, string: str) -> CapabilityStepType:
        """
        Get capability step given a string

        :param string: String to convert
        :return: Capability step corresponding to given string or Invalid
        """
        strings = {
            "prepare-and-run-workflow": cls.PrepareAndRunWorkflow,
            "await-qa": cls.AwaitQA,
            "await-workflow": cls.AwaitWorkflow,
        }
        return strings.get(string, cls.Invalid)


class CapabilityRequestState(Enum):
    """
    Enum that specifies the states that a capability request can be in
    """

    Complete = 0
    Submitted = 1
    Created = 2
    Cancelled = 3
    Failed = 4


class ExecutionState(Enum):
    """
    Enum that specifies the states that a capability execution can find itself in

    TODO: Delete this once states are fully implemented
    """

    Complete = 0
    Created = 1  # was Ready
    ExecutingWorkflow = 2
    Cancelled = 4
    Failed = -1


class QueueState(Enum):
    """
    Enum that specifies the state that an execution is in as it relates to the queue
    """

    NotQueued = 0
    Queued = 1
    Running = 2


class ExecutionPriority(Enum):
    """
    Enum that specifies the priority of a particular capability execution

    TODO: These numbers are just placeholders I added on a whim, and I think this Enum should get another look in
        the future
    """

    High = 25
    Default = 50
    Low = 100
