#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import abc
import json
import logging
import re
from typing import Iterator, List, Optional

from workspaces.capability.enums import CapabilityStepType
from workspaces.capability.helpers_interfaces import (CapabilitySequenceIF,
                                                      CapabilityStepIF,
                                                      ParameterIF)
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.interfaces import ExecutionManagerIF

logger = logging.getLogger(__name__)


class MalformedCapabilityStep(ValueError):
    pass


class CapabilityStep(CapabilityStepIF, abc.ABC):
    """
    Class that represents a step in a capability sequence. A step is of a certain type and
    certain step types can have a value
    """

    TYPES = {}

    def __init__(
        self,
        step_type: CapabilityStepType,
        step_value: Optional[str],
        step_args: Optional[str],
    ):
        self._step_type = step_type
        self._step_value = step_value
        self._step_args = step_args

    @property
    def step_type(self) -> CapabilityStepType:
        return self._step_type

    @property
    def step_value(self) -> Optional[str]:
        return self._step_value

    @property
    def step_args(self) -> Optional[str]:
        return self._step_args

    @staticmethod
    def parse_args(args_string: str) -> Optional[List[str]]:
        """
        Parse a string of the form [arg-1, arg_2, -a 3], etc. into a list of those args

        :param args_string: Args as a string
        :return: List of given args or None if no args given
        """
        args_string = args_string.replace("[", "").replace("]", "")
        return args_string.split(", ") if args_string else None

    @classmethod
    def from_str(cls, step_string: str) -> CapabilityStep:
        """
        Create CapabilityStep from string containing space-separated type, value, and args

        :param step_string: String of capability step, e.g. "PrepareAndRunWorkflow null"
        :return: CapabilityStep of given string
        """
        # === Regexes ===
        # Matches a string with letters, digits, underscores, and hyphens, labeled steptype
        # Ex: prepare-and-run-workflow, await-qa
        r_step_type = r"(?P<steptype>[\w-]+)"
        # Matches a string with letters, digits, underscores, and these specials: [./:], labeled stepval
        # Ex: cal://alma/..., workflow_name, null
        r_step_val = r"(?P<stepval>[\w\-./:]+)"
        # Matches a string surrounded by [], with 0 or more of the same string as above (plus spaces), separated by
        # ", " inside them
        # Ex: [-f, -l path/to/location, arg_value], [arg], []
        r_step_args = r"(?P<stepargs>\[(?:[\w\-./: ]+(?:, )?)*\])"
        # Matches a string taking the form of a full capability step
        # Ex: prepare-and-run-workflow null [-g], prepare-and-run-workflow download, await-qa
        r_step = rf"^{r_step_type} ?{r_step_val}? ?{r_step_args}?$"

        if match := re.search(r_step, step_string):
            groups = match.groupdict()
            step_type = groups.get("steptype", None)
            step_value = groups.get("stepval", None)
            step_args = groups.get("stepargs", None)
            if step_type:
                # Translate step_type from str to CapabilityStepType
                step_type = CapabilityStepType.from_string(step_type)

                if step_type is CapabilityStepType.Invalid:
                    raise MalformedCapabilityStep(f"Step type {step_type} invalid.")
            else:
                raise MalformedCapabilityStep(
                    f"Capability step {step_string} is malformed. Step type not found when it is required."
                )
            if step_args:
                step_args = cls.parse_args(step_args)
            return cls.TYPES[step_type](step_type, step_value, step_args)
        else:
            # Step string not well-formatted
            raise MalformedCapabilityStep(f"Capability step {step_string} is malformed.")

    def __str__(self):
        string = f"{self.step_type.name}"
        if self.step_value:
            string += f" {self.step_value}"
        elif self.step_args:
            string += f" {self.step_args}"
        return string

    @abc.abstractmethod
    def execute(self, manager: ExecutionManagerIF, execution: CapabilityExecutionIF):
        pass


class CapabilitySequence(CapabilitySequenceIF):
    """
    Class that represents a capability sequence with a list of steps
    """

    def __init__(self, steps: List[CapabilityStep]):
        self.steps = steps

    @classmethod
    def from_str(cls, sequence_str: str):
        """
        Create CapabilitySequence from comma-separated str of capability steps

        :param sequence_str: String sequence of steps
        :return: CapabilitySequence of given steps
        """
        steps = []
        for step in sequence_str.split("\n"):
            steps.append(CapabilityStep.from_str(step))

        return cls(steps)

    def __iter__(self) -> Iterator:
        """
        Allows CapabilitySequence objects to be iterated over

        :return: Iterator object
        """
        return iter(self.steps)

    def __len__(self) -> int:
        """
        Allows len() to be called on CapabilitySequence objects

        :return: Number of capability steps
        """
        return len(self.steps)

    def __getitem__(self, step_number: int) -> CapabilityStep:
        return self.steps[step_number]

    def __str__(self) -> str:
        """
        Provides a string representation of the object

        :return: String representation
        """
        str_steps = [str(step) for step in self]
        return ",".join(str_steps)


class PrepareAndRunWorkflow(CapabilityStep):
    """
    Class representing a prepare-and-run-workflow capability step
    """

    def execute(self, manager: ExecutionManagerIF, execution: CapabilityExecutionIF):
        """
        Execute a workflow within an engine

        :param manager: ExecutionManager that will execute the workflow
        :param execution: Execution record for executing capability
        """
        logger.info("Executing PrepareAndRunWorkflow for execution %s", execution.id)
        workflow_name = self.step_value
        workflow_args = None
        parameters = execution.version.parameters
        files = execution.version.files
        if parameters is not None:
            # convert to json string because json cares about single vs. double quotes.
            # DO NOT TAKE THIS OUT! Python will yell at you.
            parameters = json.dumps(parameters)
            workflow_args = json.loads(parameters)
        logger.info(
            "submitting workflow request with args %s (and type %s)",
            workflow_args,
            type(workflow_args),
        )
        manager.submit_workflow_request(execution.id, workflow_name, workflow_args, files)


class AwaitQa(CapabilityStep):
    def execute(self, manager: ExecutionManagerIF, execution: CapabilityExecutionIF):
        """
        Execute await QA step, which as of now will just send a message that triggers a notification email

        :param manager: ExecutionManager that will execute the workflow
        :param execution: Corresponding capability execution
        """
        logger.info("Executing AwaitQa on execution %s", execution.id)
        execution = manager.capability_info.lookup_execution(execution.id)
        manager.workflow_service.announce_qa(execution.current_workflow_request_id, "qa_ready")


class AwaitWorkflow(CapabilityStep):
    def execute(self, manager: ExecutionManagerIF, execution: CapabilityExecutionIF):
        logger.info("Executing AwaitWorkflow on execution %s", execution.id)


CapabilityStep.TYPES = {
    CapabilityStepType.PrepareAndRunWorkflow: PrepareAndRunWorkflow,
    CapabilityStepType.AwaitQA: AwaitQa,
    CapabilityStepType.AwaitWorkflow: AwaitWorkflow,
}


class Parameter(ParameterIF):
    def json(self) -> str:
        pass
