#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import pathlib
from typing import Dict, List, Optional

from workspaces.capability.helpers_interfaces import (CapabilityStepIF,
                                                      ParameterIF)
from workspaces.products.schema_interfaces import FutureProductIF
from workspaces.system.schema import JSONSerializable


class CapabilityIF(JSONSerializable):
    id: int
    name: str
    steps: List[CapabilityStepIF]
    max_jobs: int

    def create_request(self, parameters: ParameterIF, future_products: FutureProductIF) -> "CapabilityRequestIF":
        raise NotImplementedError


class CapabilityRequestIF(JSONSerializable):
    id: int
    state: str
    capability: CapabilityIF
    parameters: str
    future_products: str
    versions: List[CapabilityVersionIF]

    @property
    def current_version(self):
        raise NotImplementedError


class CapabilityVersionIF:
    capability_request: CapabilityRequestIF
    version_number: int
    parameters: List[ParameterIF]
    workflow_metadata: str  # JSON-formatted
    files: List[CapabilityVersionFileIF]

    def __json__(self):
        raise NotImplementedError


class CapabilityVersionFileIF:
    capability_request_id: int
    version_number: int
    filename: str
    content: bytes

    def __json__(self):
        raise NotImplementedError


class CapabilityExecutionIF:
    id: int
    capability_version: CapabilityVersionIF
    state: str
    current_step_idx: int
    current_step: CapabilityStepIF
    version: CapabilityVersionIF
    current_workflow_request_id: int
    steps: str
    capability: CapabilityIF
    capability_request: CapabilityRequestIF
    delivery_url: str
    delivery_path: pathlib.Path

    def on_last_step(self) -> bool:
        raise NotImplementedError

    def execute(self, manager: "ExecutionManagerIF"):
        raise NotImplementedError

    def __json__(self):
        raise NotImplementedError


class ActionIF:
    """
    An action to take upon performing a transition. We expect to see several implementations
    of this interface:

    - SendNotification(template, additional_args) that sends a notification with
      the event and additional arguments

    - StartWorkflow(workflow_name, additional_args) that starts a workflow with the
      provided name, the event and additional arguments
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


class PatternIF:
    def matches(self, event: dict) -> bool:
        """
        This is most likely going to be implemented as a JSON Path expression, using
        the jsonpath-python library: https://pypi.org/project/jsonpath-python/

        :param event: a JSON object
        :return: true if this pattern matches that event object
        """
        raise NotImplementedError


class StateIF:
    """
    A state that a machine could reside in.

    A state has a suite of transitions to other states. When an event comes in, we match against it; if we find a
    matching pattern, we perform that transition to another state.
    """

    transitions: List[TransitionIF]

    def link(self, pattern: str, state: "StateIf", action: ActionIF):
        raise NotImplementedError

    def signal(self, message: Dict) -> Optional["StateIf"]:
        raise NotImplementedError

    def accepting_state(self) -> bool:
        raise NotImplementedError


class TransitionIF:
    """
    A transition between states
    """

    pattern: PatternIF
    actions: List[ActionIF]
    from_state: StateIF
    to_state: StateIF

    def matches(self, msg: dict) -> bool:
        """
        True if this transition is applicable in the supplied state and matches the supplied event.

        :param msg: message whose type will be matched against
        :return: true if everything matches
        """
        raise NotImplementedError

    def action(self):
        """
        Action assodiated with particular transition

        :return:
        """
        raise NotImplementedError
