#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
# coding: utf-8
from sqlalchemy import (Column, DateTime, Float, ForeignKey, Index,
                        LargeBinary, Numeric, String, Table, Text, text)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
from sqlalchemy.sql.sqltypes import NullType

Base = declarative_base()
metadata = Base.metadata


t_SLOG_ENTRY_ATTR_bak = Table(
    "SLOG_ENTRY_ATTR_bak",
    metadata,
    Column("slog_attr_id", Numeric(19, 0, asdecimal=False), nullable=False),
    Column("slog_attr_type", Numeric(10, 0, asdecimal=False), nullable=False),
    Column("slog_attr_value", String(256)),
    Column("slog_se_id", Numeric(19, 0, asdecimal=False)),
    schema="ALMA",
)


class Aaa6(Base):
    __tablename__ = "aaa6"
    __table_args__ = {"schema": "ALMA"}

    ccc = Column(String(100), primary_key=True)


t_aaa7 = Table("aaa7", metadata, Column("ddd", String(100)), schema="ALMA")


class Account(Base):
    __tablename__ = "account"
    __table_args__ = {"schema": "ALMA"}

    account_id = Column(String(32), primary_key=True)
    request_handler_id = Column(Numeric(22, 0, asdecimal=False), unique=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    password_digest = Column(String(256))
    firstname = Column(String(256), nullable=False)
    lastname = Column(String(256), nullable=False)
    initials = Column(String(256))
    creationtimestamp = Column(String(32))
    modificationtimestamp = Column(String(32))
    preferredarc = Column(String(32))
    email = Column(String(256), unique=True)
    executive = Column(String(5))
    aliases = Column(String(256))
    inst_no = Column(ForeignKey("ALMA.institution.inst_no"), nullable=False)
    active = Column(String(1), nullable=False, server_default=text("'T' "))
    account_id_merged_to = Column(ForeignKey("ALMA.account.account_id"))
    merge_time = Column(DateTime)
    receiveemails = Column(String(1), nullable=False)
    passwordtimestamp = Column(DateTime)
    provides_user_demographics = Column(String(1))

    parent = relationship("Account", remote_side=[account_id])
    institution = relationship("Institution")
    nodes = relationship("DrmNode", secondary="ALMA.drm_data_reducer")
    role = relationship("Role", secondary="ALMA.account_role")


class AccountDelegation(Base):
    __tablename__ = "account_delegation"
    __table_args__ = (
        Index("account_del_unique", "pi_rh_id", "project_code", "delegee_rh_id", "delegation_type", unique=True),
        {"schema": "ALMA"},
    )

    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    pi_rh_id = Column(ForeignKey("ALMA.account.request_handler_id"), nullable=False)
    project_code = Column(String(18), nullable=False)
    delegee_rh_id = Column(ForeignKey("ALMA.account.request_handler_id"), nullable=False)
    delegation_type = Column(String(10), nullable=False, index=True, server_default=text("'DATA'                "))

    delegee_rh = relationship("Account", primaryjoin="AccountDelegation.delegee_rh_id == Account.request_handler_id")
    pi_rh = relationship("Account", primaryjoin="AccountDelegation.pi_rh_id == Account.request_handler_id")


t_account_role = Table(
    "account_role",
    metadata,
    Column("role_no", ForeignKey("ALMA.role.role_no"), primary_key=True, nullable=False),
    Column("account_id", ForeignKey("ALMA.account.account_id"), primary_key=True, nullable=False),
    schema="ALMA",
)


t_alma_alarms = Table(
    "alma_alarms",
    metadata,
    Column("descriptor", Numeric(10, 0, asdecimal=False)),
    Column("user_timestamp_secs", Numeric(20, 0, asdecimal=False)),
    Column("user_timestamp_ms", Numeric(20, 0, asdecimal=False)),
    Column("activated_by_backup", Numeric(1, 0, asdecimal=False)),
    Column("terminated_by_backup", Numeric(1, 0, asdecimal=False)),
    Column("family", String(100)),
    Column("member", String(100)),
    Column("code", Numeric(10, 0, asdecimal=False)),
    Column("user_properties", String(1000)),
    Column("source_name", String(100)),
    Column("source_hostname", String(100)),
    Column("source_timestamp_sec", Numeric(20, 0, asdecimal=False)),
    Column("source_timestamp_ms", Numeric(20, 0, asdecimal=False)),
    Column("backup", Numeric(1, 0, asdecimal=False)),
    Column("version", String(100)),
    schema="ALMA",
)


class AquaAttachment(Base):
    __tablename__ = "aqua_attachment"
    __table_args__ = (Index("aqua_attachment_uni", "name", "ttimestamp", unique=True), {"schema": "ALMA"})

    attachmentid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    ttimestamp = Column(DateTime)
    name = Column(String(256), nullable=False)
    contents = Column(LargeBinary, nullable=False)

    aqua_session = relationship("AquaSession", secondary="ALMA.aqua_session_attachment")
    aqua_execblock = relationship("AquaExecblock", secondary="ALMA.aqua_execblock_attachment")


class AquaComment(Base):
    __tablename__ = "aqua_comment"
    __table_args__ = (Index("aqua_comment_uni", "author", "ttimestamp", unique=True), {"schema": "ALMA"})

    commentid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    author = Column(String(32), nullable=False)
    ttimestamp = Column(DateTime, nullable=False)
    ccomment = Column(Text)

    aqua_session = relationship("AquaSession", secondary="ALMA.aqua_session_comment")
    aqua_execblock = relationship("AquaExecblock", secondary="ALMA.aqua_execblock_comment")


class AquaExecblock(Base):
    __tablename__ = "aqua_execblock"
    __table_args__ = {"schema": "ALMA"}

    execblockid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    execblockuid = Column(String(32), nullable=False, unique=True)
    sessionid = Column(ForeignKey("ALMA.aqua_session.sessionid"), nullable=False)
    qa0status = Column(String(32), nullable=False, index=True)
    starttime = Column(DateTime)
    endtime = Column(DateTime)
    finalcommentid = Column(ForeignKey("ALMA.aqua_comment.commentid"), unique=True)
    schedblockuid = Column(String(32))
    obsprojectcode = Column(String(25))
    obsprojectname = Column(String(256))
    obsprojectuid = Column(String(32))
    location = Column(String(32))
    correlatortype = Column(String(4))
    execfraction = Column(Numeric(5, 3), nullable=False, server_default=text("0 "))

    aqua_comment = relationship("AquaComment")
    aqua_session = relationship("AquaSession")


t_aqua_execblock_attachment = Table(
    "aqua_execblock_attachment",
    metadata,
    Column("attachmentid", ForeignKey("ALMA.aqua_attachment.attachmentid"), primary_key=True),
    Column("execblockid", ForeignKey("ALMA.aqua_execblock.execblockid"), nullable=False),
    schema="ALMA",
)


t_aqua_execblock_comment = Table(
    "aqua_execblock_comment",
    metadata,
    Column("commentid", ForeignKey("ALMA.aqua_comment.commentid"), primary_key=True),
    Column("execblockid", ForeignKey("ALMA.aqua_execblock.execblockid")),
    schema="ALMA",
)


class AquaExecblockEtlDatum(Base):
    __tablename__ = "aqua_execblock_etl_data"
    __table_args__ = (Index("eb_type_unique", "execblockid", "etl_data_type", unique=True), {"schema": "ALMA"})

    etl_data_id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    execblockid = Column(ForeignKey("ALMA.aqua_execblock.execblockid"))
    etl_data_type = Column(String(16))
    etl_schema_version = Column(Numeric(3, 0, asdecimal=False))
    etl_data = Column(Text)

    aqua_execblock = relationship("AquaExecblock")


class AquaOu(Base):
    __tablename__ = "aqua_ous"
    __table_args__ = (Index("aqua_ous_uni", "obsunitsetuid", "obsunitsetpartid", unique=True), {"schema": "ALMA"})

    obsunitsetid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    obsunitsetuid = Column(String(32), nullable=False)
    qa2status = Column(String(8), nullable=False)
    finalcommentid = Column(ForeignKey("ALMA.aqua_comment.commentid"), unique=True)
    obsunitsetpartid = Column(String(32), nullable=False)
    qa2semipassreason = Column(String(11))
    qa2statustime = Column(DateTime)
    achievedsens = Column(Numeric(6, 2))
    achievedresol = Column(Numeric(6, 2))
    qa2plcalstatus = Column(String(8))
    qa2mancalstatus = Column(String(8))
    plcalfinalcommentid = Column(Numeric(10, 0, asdecimal=False))
    mcalfinalcommentid = Column(Numeric(10, 0, asdecimal=False))
    casatickets = Column(String(128))
    analyst = Column(String(32))
    execfraction = Column(Numeric(5, 3), nullable=False, server_default=text("0 "))
    plimgfinalcommentid = Column(Numeric(10, 0, asdecimal=False))
    mimgfinalcommentid = Column(Numeric(10, 0, asdecimal=False))
    rmsatbandwidth = Column(NullType)
    rmscontinuum = Column(NullType)
    beamsemimajor = Column(NullType)
    beamsemiminor = Column(NullType)
    beamposangle = Column(NullType)
    ous_status_entity_id = Column(String(64))

    aqua_comment = relationship("AquaComment")


class AquaOusAttachment(Base):
    __tablename__ = "aqua_ous_attachment"
    __table_args__ = {"schema": "ALMA"}

    attachmentid = Column(ForeignKey("ALMA.aqua_attachment.attachmentid"), primary_key=True, nullable=False)
    obsunitsetid = Column(ForeignKey("ALMA.aqua_ous.obsunitsetid"), primary_key=True, nullable=False)
    attachment_section = Column(String(8))

    aqua_attachment = relationship("AquaAttachment")
    aqua_ou = relationship("AquaOu")


class AquaOusFlag(Base):
    __tablename__ = "aqua_ous_flag"
    __table_args__ = (
        Index("unique_ous_flag", "ous_status_entity_id", "flag_name", unique=True),
        Index("idx_aqua_ous_flag_val", "flag_name", "flag_value"),
        {"schema": "ALMA"},
    )

    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(4, 0, asdecimal=False), nullable=False)
    ous_status_entity_id = Column(ForeignKey("ALMA.obs_unit_set_status.status_entity_id"), nullable=False, index=True)
    flag_name = Column(String(32), nullable=False)
    flag_value = Column(String(2000))
    update_account_id = Column(ForeignKey("ALMA.account.account_id"), nullable=False)
    update_timestamp = Column(DateTime, nullable=False)

    ous_status_entity = relationship("ObsUnitSetStatu")
    update_account = relationship("Account")


class AquaOusFlagHistory(Base):
    __tablename__ = "aqua_ous_flag_history"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    ous_status_entity_id = Column(ForeignKey("ALMA.obs_unit_set_status.status_entity_id"), nullable=False)
    flag_name = Column(String(32), nullable=False)
    flag_value = Column(String(2000))
    update_account_id = Column(ForeignKey("ALMA.account.account_id"), nullable=False)
    update_timestamp = Column(DateTime, nullable=False)

    ous_status_entity = relationship("ObsUnitSetStatu")
    update_account = relationship("Account")


class AquaPipelineRun(Base):
    __tablename__ = "aqua_pipeline_run"
    __table_args__ = (Index("uk_mous_uid_timestamp", "mous_uid", "timestamp", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    gous_uid = Column(String(32))
    mous_uid = Column(String(32))
    report = Column(Text)
    sous_uid = Column(String(32))
    timestamp = Column(String(32))
    version = Column(Numeric(10, 0, asdecimal=False))
    pipeline_run_dir_id = Column(ForeignKey("ALMA.aqua_pipeline_run_directory.id"))
    weblog_id = Column(ForeignKey("ALMA.aqua_weblog.id"))

    pipeline_run_dir = relationship("AquaPipelineRunDirectory")
    weblog = relationship("AquaWeblog")


class AquaPipelineRunDirectory(Base):
    __tablename__ = "aqua_pipeline_run_directory"
    __table_args__ = (Index("uk_project_code_timestamp", "project_code", "timestamp", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    remark = Column(String(2000))
    last_op_timestamp = Column(String(32))
    pathname = Column(String(512))
    project_code = Column(String(32))
    status = Column(String(32))
    timestamp = Column(String(32))
    version = Column(Numeric(10, 0, asdecimal=False))
    last_op_account = Column(String(32))
    products_sub_dir = Column(String(512))
    region = Column(String(32))
    upload_timestamp = Column(String(32))
    data_reduction_type = Column(String(32))


class AquaQa2EcHistory(Base):
    __tablename__ = "aqua_qa2_ec_history"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    ttimestamp = Column(DateTime, nullable=False)
    entityuid = Column(String(32), nullable=False)
    author = Column(String(32), nullable=False)
    ec = Column(NullType)


class AquaQa2FailReason(Base):
    __tablename__ = "aqua_qa2_fail_reason"
    __table_args__ = {"schema": "ALMA"}

    qa2reasonid = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    obsunitsetid = Column(ForeignKey("ALMA.aqua_ous.obsunitsetid"), nullable=False)
    qa2failreason = Column(String(20), nullable=False)

    aqua_ou = relationship("AquaOu")


class AquaQa2ImagingHistory(Base):
    __tablename__ = "aqua_qa2_imaging_history"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    ttimestamp = Column(DateTime, nullable=False)
    entityuid = Column(String(32), nullable=False)
    author = Column(String(32), nullable=False)
    rmsatbandwidth = Column(NullType)
    rmscontinuum = Column(NullType)
    beamsemimajor = Column(NullType)
    beamsemiminor = Column(NullType)
    beamposangle = Column(NullType)


class AquaSensitivity(Base):
    __tablename__ = "aqua_sensitivity"
    __table_args__ = {"schema": "ALMA"}

    asdmuid = Column(String(32), primary_key=True, nullable=False)
    spwid = Column(Numeric(38, 0, asdecimal=False), primary_key=True, nullable=False)
    mediantsys = Column(NullType)


class AquaSession(Base):
    __tablename__ = "aqua_session"
    __table_args__ = (Index("aqua_session_uni", "sessionuid", "sessionpartid", unique=True), {"schema": "ALMA"})

    sessionid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    sessionuid = Column(String(32), nullable=False)
    schedblockuid = Column(String(32), nullable=False)
    schedblockname = Column(String(128))
    projectcode = Column(String(64))
    qlfocussummuid = Column(String(32))
    qlpointsummuid = Column(String(32))
    qlphasesummuid = Column(String(32))
    qlatmsummuid = Column(String(32))
    qa1status = Column(String(5), nullable=False)
    starttime = Column(DateTime, nullable=False)
    endtime = Column(DateTime)
    finalcommentid = Column(ForeignKey("ALMA.aqua_comment.commentid"), unique=True)
    sessionpartid = Column(String(32), nullable=False)
    location = Column(String(32))
    tmcdbconfigurationname = Column(String(128))
    correlatortype = Column(String(4))

    aqua_comment = relationship("AquaComment")


t_aqua_session_attachment = Table(
    "aqua_session_attachment",
    metadata,
    Column("attachmentid", ForeignKey("ALMA.aqua_attachment.attachmentid"), primary_key=True),
    Column("sessionid", ForeignKey("ALMA.aqua_session.sessionid"), nullable=False),
    schema="ALMA",
)


t_aqua_session_comment = Table(
    "aqua_session_comment",
    metadata,
    Column("commentid", ForeignKey("ALMA.aqua_comment.commentid"), primary_key=True),
    Column("sessionid", ForeignKey("ALMA.aqua_session.sessionid")),
    schema="ALMA",
)


class AquaStatusHistory(Base):
    __tablename__ = "aqua_status_history"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    entityuid = Column(String(32), nullable=False)
    author = Column(String(32), nullable=False)
    ttimestamp = Column(DateTime, nullable=False)
    qastatus = Column(String(32), nullable=False)
    qastatustype = Column(String(32), nullable=False)
    ccomment = Column(Text)


t_aqua_v_execblock = Table(
    "aqua_v_execblock",
    metadata,
    Column("execblockid", Numeric(10, 0, asdecimal=False), nullable=False),
    Column("execblockuid", String(32), nullable=False),
    Column("execfraction", Numeric(5, 3), nullable=False),
    Column("qa0status", String(32), nullable=False),
    Column("finalcommentid", Numeric(10, 0, asdecimal=False)),
    Column("sessionid", Numeric(10, 0, asdecimal=False), nullable=False),
    Column("sessionuid", String(32)),
    Column("sessionpartid", String(32)),
    Column("starttime", DateTime),
    Column("endtime", DateTime),
    Column("schedblockuid", String(32)),
    Column("schedblockname", String(128)),
    Column("obsprojectcode", String(128)),
    Column("obsprojectname", String(256)),
    Column("obsprojectpi", String(20)),
    Column("obsprojectversion", String(20)),
    Column("location", String(32)),
    Column("status", String(32)),
    Column("archivingstatus", String(32)),
    Column("bandname", String(10)),
    Column("representativefrequency", Numeric(15, 10)),
    Column("almabuild", String(70)),
    Column("arrayname", String(10)),
    Column("arraystart", DateTime),
    Column("arrayend", DateTime),
    Column("correlatortype", String(4)),
    Column("arrayfamily", String(11)),
    Column("arraytype", String(9)),
    Column("photonicreferencename", String(18)),
    Column("obsprojectuid", String(32)),
    schema="ALMA",
)


class AquaWeblog(Base):
    __tablename__ = "aqua_weblog"
    __table_args__ = (Index("uk_ous_uid_timestamp", "ous_uid", "timestamp", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    cached = Column(String(1))
    last_access = Column(String(32))
    ous_uid = Column(String(32))
    timestamp = Column(String(32))
    version = Column(Numeric(10, 0, asdecimal=False))
    last_op_account = Column(String(32))
    remark = Column(String(2000))
    last_op_timestamp = Column(String(32))
    ngas_file_id = Column(String(128))
    status = Column(String(32))


class AsaBibliography(Base):
    __tablename__ = "asa_bibliography"
    __table_args__ = {"schema": "ALMA"}

    bibcode = Column(String(30), nullable=False, unique=True)
    publication_year = Column(Numeric(4, 0, asdecimal=False))
    journal = Column(String(100), nullable=False)
    title = Column(String(255), nullable=False, index=True)
    first_author = Column(String(255), nullable=False)
    volume = Column(Numeric(8, 0, asdecimal=False))
    pages = Column(String(100))
    abstract = Column(String(4000))
    authors = Column(String(4000))
    keywords = Column(String(4000))
    bib_id = Column(Numeric(30, 0, asdecimal=False), primary_key=True)


t_asa_bibliography_temp = Table(
    "asa_bibliography_temp",
    metadata,
    Column("bibcode", String(30), nullable=False),
    Column("publication_year", Numeric(4, 0, asdecimal=False)),
    Column("journal", String(100), nullable=False),
    Column("title", String(255), nullable=False),
    Column("first_author", String(255), nullable=False),
    Column("volume", Numeric(8, 0, asdecimal=False)),
    Column("pages", String(100)),
    Column("abstract", String(4000)),
    Column("authors", String(4000)),
    Column("keywords", String(4000)),
    Column("bib_id", Numeric(30, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


class AsaColumn(Base):
    __tablename__ = "asa_columns"
    __table_args__ = (Index("asa_columns_unique", "table_name", "column_name", unique=True), {"schema": "ALMA"})

    column_name = Column(String(128), nullable=False)
    table_name = Column(String(128), nullable=False)
    description = Column(String(1024))
    tooltip = Column(String(256))
    unit = Column(String(32))
    dimensions = Column(String(32))
    scale = Column(Float)
    ucd = Column(String(128))
    utype = Column(String(512))
    datatype = Column(String(32))
    type_size = Column(Numeric(11, 0, asdecimal=False))
    principal = Column(Numeric(1, 0, asdecimal=False), nullable=False)
    indexed = Column(Numeric(1, 0, asdecimal=False), nullable=False)
    std = Column(Numeric(1, 0, asdecimal=False), nullable=False)
    is_enum = Column(Numeric(1, 0, asdecimal=False), nullable=False)
    enum_table = Column(String(128))
    examples = Column(String(1024))
    is_queriable = Column(String(1), nullable=False)
    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)


class AsaDeliveryAsdmOu(Base):
    __tablename__ = "asa_delivery_asdm_ous"
    __table_args__ = (Index("dp_delao_ad_unq", "asdm_uid", "deliverable_name", unique=True), {"schema": "ALMA"})

    asdm_uid = Column(ForeignKey("ALMA.xml_asdm_entities.archive_uid"), index=True)
    ous_status_id = Column(String(33), index=True)
    deliverable_name = Column(String(64), nullable=False, index=True)
    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    delivery_id = Column(ForeignKey("ALMA.asa_delivery_status.id"), nullable=False)

    xml_asdm_entity = relationship("XmlAsdmEntity")
    delivery = relationship("AsaDeliveryStatu")


class AsaDeliveryStatu(Base):
    __tablename__ = "asa_delivery_status"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    delivery_id = Column(String(64), nullable=False, unique=True)
    pi_account_id = Column(Numeric(22, 0, asdecimal=False), nullable=False)
    qa2_passed = Column(String(1), nullable=False)
    notification_date = Column(DateTime)
    original_release_date = Column(DateTime)
    release_date = Column(DateTime, server_default=text("to_date('3000-01-01', 'YYYY-MM-DD')"))
    release_date_comment = Column(String(4000))
    delivery_name = Column(String(64), nullable=False)


class AsaEnergy(Base):
    __tablename__ = "asa_energy"
    __table_args__ = {"schema": "ALMA"}

    asa_energy_id = Column(String(128), primary_key=True)
    asa_dataset_id = Column(ForeignKey("ALMA.asa_science.dataset_id"), index=True)
    frequency_min = Column(NullType, nullable=False)
    frequency_max = Column(NullType, nullable=False)
    resolution_min = Column(NullType, nullable=False)
    resolution_max = Column(NullType, nullable=False)
    channel_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    channel_width_min = Column(NullType, nullable=False)
    channel_width_max = Column(NullType, nullable=False)
    resolving_power_min = Column(NullType, nullable=False)
    resolving_power_max = Column(NullType, nullable=False)
    spectral_window_uid = Column(String(64))
    bandwidth = Column(NullType)
    pol_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    pol_product = Column(String(64), nullable=False)
    exptimtp = Column(NullType)
    exptim7 = Column(NullType)
    exptim12 = Column(NullType)
    restfreq = Column(NullType)
    chanrms = Column(NullType)
    crval3 = Column(NullType)
    crpix3 = Column(NullType)
    crval4 = Column(NullType)
    crpix4 = Column(NullType)
    cdelt4 = Column(NullType)
    parent_asa_energy_ids = Column(String(256))
    spw_num = Column(Numeric(38, 0, asdecimal=False))
    spw_id_original = Column(Numeric(38, 0, asdecimal=False))
    spw_id_orig_list = Column(String(64))
    spectral_window_name = Column(String(64))
    asa_energy_product_id = Column(String(128))
    asdm_spw_id = Column(Numeric(38, 0, asdecimal=False))
    ms_spw_id = Column(Numeric(38, 0, asdecimal=False))
    sensitivity_channel = Column(NullType)
    sensitivity_10kms = Column(NullType)
    sensitivity_bandwidth = Column(NullType)
    line_id = Column(String(256))
    rest_frequency_max = Column(NullType)
    rest_frequency_min = Column(NullType)
    rest_frequency_resolution_min = Column(NullType)
    rest_frequency_resolution_max = Column(NullType)
    rest_resolving_power_min = Column(NullType)
    rest_resolving_power_max = Column(NullType)
    rest_channel_width_min = Column(NullType)
    rest_channel_width_max = Column(NullType)
    rest_bandwidth = Column(NullType)
    parent_energy_id = Column(String(128))

    asa_dataset = relationship("AsaScience")


class AsaEnergyTemp(Base):
    __tablename__ = "asa_energy_temp"
    __table_args__ = {"schema": "ALMA"}

    asa_energy_id = Column(String(128), primary_key=True)
    asa_dataset_id = Column(String(128), index=True)
    frequency_min = Column(NullType, nullable=False)
    frequency_max = Column(NullType, nullable=False)
    resolution_min = Column(NullType, nullable=False)
    resolution_max = Column(NullType, nullable=False)
    channel_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    channel_width_min = Column(NullType, nullable=False)
    channel_width_max = Column(NullType, nullable=False)
    resolving_power_min = Column(NullType, nullable=False)
    resolving_power_max = Column(NullType, nullable=False)
    spectral_window_uid = Column(String(64))
    bandwidth = Column(NullType)
    pol_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    pol_product = Column(String(64), nullable=False)
    exptimtp = Column(NullType)
    exptim7 = Column(NullType)
    exptim12 = Column(NullType)
    restfreq = Column(NullType)
    chanrms = Column(NullType)
    crval3 = Column(NullType)
    crpix3 = Column(NullType)
    crval4 = Column(NullType)
    crpix4 = Column(NullType)
    cdelt4 = Column(NullType)
    parent_asa_energy_ids = Column(String(256))
    spw_num = Column(Numeric(38, 0, asdecimal=False))
    spw_id_original = Column(Numeric(38, 0, asdecimal=False))
    spw_id_orig_list = Column(String(64))
    spectral_window_name = Column(String(64))
    asa_energy_product_id = Column(String(128), index=True)
    asdm_spw_id = Column(Numeric(38, 0, asdecimal=False), index=True)
    ms_spw_id = Column(Numeric(38, 0, asdecimal=False))
    sensitivity_channel = Column(NullType)
    sensitivity_10kms = Column(NullType)
    sensitivity_bandwidth = Column(NullType)
    line_id = Column(String(256))
    rest_frequency_max = Column(NullType)
    rest_frequency_min = Column(NullType)
    rest_frequency_resolution_min = Column(NullType)
    rest_frequency_resolution_max = Column(NullType)
    rest_resolving_power_min = Column(NullType)
    rest_resolving_power_max = Column(NullType)
    rest_channel_width_min = Column(NullType)
    rest_channel_width_max = Column(NullType)
    rest_bandwidth = Column(NullType)
    parent_energy_id = Column(String(128))


class AsaMailTemplate(Base):
    __tablename__ = "asa_mail_templates"
    __table_args__ = {"schema": "ALMA"}

    templatename = Column(String(100), primary_key=True, nullable=False)
    application = Column(String(100), primary_key=True, nullable=False)
    templatecontent = Column(Text, nullable=False)


class AsaOu(Base):
    __tablename__ = "asa_ous"
    __table_args__ = {"schema": "ALMA"}

    asa_ous_id = Column(String(64), primary_key=True)
    asa_project_code = Column(String(64))
    sg_ous_uid = Column(String(64))
    group_ous_uid = Column(String(64))
    member_ous_uid = Column(String(64))
    member_num = Column(Numeric(10, 0, asdecimal=False))
    completed = Column(Numeric(1, 0, asdecimal=False))
    xml_scipiperesults_archive_uid = Column(ForeignKey("ALMA.xml_scipiperesults_entities.archive_uid"))
    individual_files = Column(Numeric(1, 0, asdecimal=False), nullable=False, server_default=text("0 "))

    xml_scipiperesults_entity = relationship("XmlScipiperesultsEntity")


class AsaOusRedshift(Base):
    __tablename__ = "asa_ous_redshift"
    __table_args__ = {"schema": "ALMA"}

    group_ous_uid = Column(String(64), primary_key=True, nullable=False)
    member_ous_uid = Column(String(64), primary_key=True, nullable=False)
    source_name = Column(String(256), primary_key=True, nullable=False)
    best_source_redshift = Column(NullType)


class AsaOusTemp(Base):
    __tablename__ = "asa_ous_temp"
    __table_args__ = {"schema": "ALMA"}

    asa_ous_id = Column(String(64), primary_key=True)
    asa_project_code = Column(String(64))
    sg_ous_uid = Column(String(64))
    group_ous_uid = Column(String(64))
    member_ous_uid = Column(String(64))
    member_num = Column(Numeric(10, 0, asdecimal=False))
    completed = Column(Numeric(1, 0, asdecimal=False))
    xml_scipiperesults_archive_uid = Column(String(33))
    individual_files = Column(Numeric(1, 0, asdecimal=False), nullable=False)


class AsaProductFile(Base):
    __tablename__ = "asa_product_files"
    __table_args__ = {"schema": "ALMA"}

    ngas_file_id = Column(String(220), primary_key=True, nullable=False)
    asa_ous_id = Column(ForeignKey("ALMA.asa_ous.asa_ous_id"), primary_key=True, nullable=False)
    asa_science_id = Column(ForeignKey("ALMA.asa_science.dataset_id"))
    asa_energy_id = Column(ForeignKey("ALMA.asa_energy.asa_energy_id"))
    stored_size = Column(Numeric(20, 0, asdecimal=False), nullable=False)
    subdirectory = Column(String(128), nullable=False)
    file_class = Column(String(64))
    file_type = Column(String(64))
    pack = Column(Numeric(1, 0, asdecimal=False), nullable=False)
    vo = Column(Numeric(1, 0, asdecimal=False))
    creation_date = Column(DateTime, nullable=False)
    naxis = Column(Numeric(38, 0, asdecimal=False))
    naxis1 = Column(Numeric(38, 0, asdecimal=False))
    naxis2 = Column(Numeric(38, 0, asdecimal=False))
    naxis3 = Column(Numeric(38, 0, asdecimal=False))
    naxis4 = Column(Numeric(38, 0, asdecimal=False))
    parent_product_files_id = Column(String(64))
    main_product_source = Column(Numeric(1, 0, asdecimal=False))
    main_product_spw = Column(Numeric(1, 0, asdecimal=False))
    origin = Column(String(32))
    user_id = Column(String(32))
    processing_recipe = Column(String(7), nullable=False, server_default=text("'CAL+IMG' "))

    asa_energy = relationship("AsaEnergy")
    asa_ous = relationship("AsaOu")
    asa_science = relationship("AsaScience")


class AsaProject(Base):
    __tablename__ = "asa_project"
    __table_args__ = {"schema": "ALMA"}

    code = Column(String(20), primary_key=True)
    project_uid = Column(String(32), nullable=False, unique=True)
    proposal_uid = Column(ForeignKey("ALMA.xml_obsproposal_entities.archive_uid"), nullable=False)
    pi_name = Column(String(256), nullable=False)
    pi_userid = Column(ForeignKey("ALMA.account.account_id"), nullable=False)
    coi_name = Column(String(2000))
    title = Column(String(256))
    type = Column(String(16), nullable=False)
    associated_arc = Column(String(8), nullable=False)
    proposal_cycle = Column(String(6), nullable=False)
    proposed_start_date = Column(String(128))
    proposed_end_date = Column(String(128))
    first_obs_start = Column(String(128))
    first_obs_end = Column(String(128))
    last_obs_start = Column(String(128))
    last_obs_end = Column(String(128))
    arc_release = Column(String(128))
    scientific_category = Column(String(200))
    science_keyword = Column(String(200))
    proposal_abstract = Column(String(4000))

    account = relationship("Account")
    xml_obsproposal_entity = relationship("XmlObsproposalEntity")

    def __repr__(self):
        return '<AsaProject#{code} "{title}" start={first_obs_start} end={last_obs_end}>'.format(**self.__dict__)


class AsaProjectBibliography(Base):
    __tablename__ = "asa_project_bibliography"
    __table_args__ = (Index("bibprojectcode_pk", "bibcode", "project_code", unique=True), {"schema": "ALMA"})

    bibcode = Column(String(30), nullable=False)
    project_code = Column(ForeignKey("ALMA.asa_project.code"), primary_key=True, nullable=False)
    bib_id = Column(ForeignKey("ALMA.asa_bibliography.bib_id"), primary_key=True, nullable=False)

    bib = relationship("AsaBibliography")
    asa_project = relationship("AsaProject")


t_asa_project_bibliography_temp = Table(
    "asa_project_bibliography_temp",
    metadata,
    Column("bibcode", String(30), nullable=False),
    Column("project_code", String(20), nullable=False),
    Column("bib_id", Numeric(30, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


class AsaProjectTemp(Base):
    __tablename__ = "asa_project_temp"
    __table_args__ = {"schema": "ALMA"}

    code = Column(String(20), primary_key=True)
    project_uid = Column(String(32), nullable=False)
    proposal_uid = Column(String(32), nullable=False)
    pi_name = Column(String(256), nullable=False)
    pi_userid = Column(String(256), nullable=False)
    coi_name = Column(String(2000))
    title = Column(String(256))
    type = Column(String(16), nullable=False)
    associated_arc = Column(String(8), nullable=False)
    proposal_cycle = Column(String(6), nullable=False)
    proposed_start_date = Column(String(128))
    proposed_end_date = Column(String(128))
    first_obs_start = Column(String(128))
    first_obs_end = Column(String(128))
    last_obs_start = Column(String(128))
    last_obs_end = Column(String(128))
    arc_release = Column(String(128))
    scientific_category = Column(String(200))
    science_keyword = Column(String(200))
    proposal_abstract = Column(String(4000))


class AsaScience(Base):
    __tablename__ = "asa_science"
    __table_args__ = {"schema": "ALMA"}

    dataset_id = Column(String(128), primary_key=True)
    asdm_uid = Column(String(32), index=True)
    field_uid = Column(String(32))
    spectral_window_uid = Column(String(32))
    project_uid = Column(ForeignKey("ALMA.asa_project.project_uid"))
    schedblock_uid = Column(String(32))
    execblock_uid = Column(String(32))
    ra = Column(NullType, nullable=False)
    dec = Column(NullType, nullable=False)
    ra_source = Column(NullType, nullable=False)
    dec_source = Column(NullType, nullable=False)
    cx = Column(NullType)
    cy = Column(NullType)
    cz = Column(NullType)
    az_min = Column(NullType, nullable=False)
    az_max = Column(NullType, nullable=False)
    elevation_min = Column(NullType, nullable=False)
    elevation_max = Column(NullType, nullable=False)
    wcs_cd1_1 = Column(NullType)
    wcs_cd1_2 = Column(NullType)
    wcs_cd2_1 = Column(NullType)
    wcs_cd2_2 = Column(NullType)
    wcs_crpix1 = Column(NullType)
    wcs_crpix2 = Column(NullType)
    wcs_crval1 = Column(NullType)
    wcs_crval2 = Column(NullType)
    spatial_scale_min = Column(NullType, nullable=False)
    spatial_scale_max = Column(NullType, nullable=False)
    gal_longitude = Column(NullType, nullable=False)
    gal_latitude = Column(NullType, nullable=False)
    ecliptic_longitude = Column(NullType, nullable=False)
    ecliptic_latitude = Column(NullType, nullable=False)
    footprint = Column(String(2048))
    fov = Column(NullType, nullable=False)
    bounding_box = Column(String(2048))
    source_name = Column(String(256))
    source_is_in_simbad = Column(String(1), nullable=False, server_default=text("'N' "))
    source_class = Column(String(128))
    source_redshift = Column(NullType)
    source_alma_catalog_id = Column(Numeric(22, 0, asdecimal=False))
    start_date = Column(DateTime, nullable=False)
    end_date = Column(DateTime, nullable=False)
    int_time = Column(Numeric(10, 3), nullable=False)
    time_resolution = Column(Numeric(8, 4), nullable=False)
    frequency = Column(NullType)
    frequency_resolution = Column(NullType)
    velocity_resolution = Column(NullType)
    resolving_power = Column(NullType)
    bandwidth = Column(NullType)
    rest_frequency = Column(NullType)
    rest_frequency_min = Column(NullType)
    rest_frequency_max = Column(NullType)
    rest_frequency_resolution = Column(NullType)
    rest_velocity_resolution = Column(NullType, nullable=False)
    rest_resolving_power = Column(NullType, nullable=False)
    rest_frequency_support = Column(String(512), nullable=False)
    channel_num = Column(Numeric(scale=0, asdecimal=False))
    spectral_window_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    continuum_window_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    obs_unitset_id = Column(String(32), nullable=False)
    science_goal_ouss_id = Column(ForeignKey("ALMA.obs_unit_set_status.status_entity_id"))
    group_ouss_id = Column(ForeignKey("ALMA.obs_unit_set_status.status_entity_id"))
    member_ouss_id = Column(ForeignKey("ALMA.obs_unit_set_status.status_entity_id"))
    sensitivity = Column(Numeric(22, 0, asdecimal=False))
    flux_min = Column(NullType)
    flux_max = Column(NullType)
    uv_coverage_param1 = Column(NullType)
    uv_coverage_param2 = Column(NullType)
    pol_num = Column(Numeric(scale=0, asdecimal=False))
    pol_products = Column(String(64))
    airmass = Column(NullType)
    pwv = Column(NullType)
    temperature = Column(NullType)
    windspeed = Column(NullType)
    winddirection = Column(NullType)
    antennas = Column(String(660))
    baseline_max = Column(NullType, nullable=False)
    baseline_min = Column(NullType, nullable=False)
    scan_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    project_code = Column(ForeignKey("ALMA.asa_project.code"), nullable=False, index=True)
    schedblock_name = Column(String(128))
    observation_category = Column(String(128), nullable=False)
    scan_intent = Column(String(256))
    access_format = Column(String(10), nullable=False)
    processing_level = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    frequency_support = Column(String(4000))
    spatial_resolution = Column(NullType, nullable=False, server_default=text("-1.0 "))
    spatres = Column(NullType)
    mineltp = Column(NullType)
    minel12 = Column(NullType)
    minel7 = Column(NullType)
    max_angscale = Column(NullType)
    obs_mode = Column(String(80))
    pipver = Column(String(80))
    asdm_list = Column(String(200))
    ant_tp_num = Column(Numeric(3, 0, asdecimal=False), nullable=False, server_default=text("0 "))
    ant_num = Column(Numeric(3, 0, asdecimal=False), nullable=False, server_default=text("0 "))
    ant_main_num = Column(Numeric(3, 0, asdecimal=False), nullable=False, server_default=text("0 "))
    ant_aca_num = Column(Numeric(3, 0, asdecimal=False), nullable=False, server_default=text("0 "))
    frequency_max = Column(NullType, nullable=False, server_default=text("-1. "))
    frequency_min = Column(NullType, nullable=False, server_default=text("-1. "))
    requestedarray = Column(String(8))
    asa_ous_id = Column(ForeignKey("ALMA.asa_ous.asa_ous_id"))
    band_list = Column(String(30))
    antenna_string = Column(String(30))
    product_type = Column(String(10))
    dataset_product_id = Column(String(128))
    rms = Column(NullType)
    is_mosaic = Column(String(1), nullable=False, server_default=text("'N' "))
    best_source_redshift = Column(NullType)
    redshift_provenance = Column(String(1))
    cont_sensitivity_channel = Column(NullType)
    cont_sensitivity_10kms = Column(NullType)
    cont_sensitivity_bandwidth = Column(NullType)
    parent_dataset_id = Column(String(128))

    asa_ous = relationship("AsaOu")
    group_ouss = relationship(
        "ObsUnitSetStatu", primaryjoin="AsaScience.group_ouss_id == ObsUnitSetStatu.status_entity_id"
    )
    member_ouss = relationship(
        "ObsUnitSetStatu", primaryjoin="AsaScience.member_ouss_id == ObsUnitSetStatu.status_entity_id"
    )
    asa_project = relationship("AsaProject", primaryjoin="AsaScience.project_code == AsaProject.code")
    asa_project1 = relationship("AsaProject", primaryjoin="AsaScience.project_uid == AsaProject.project_uid")
    science_goal_ouss = relationship(
        "ObsUnitSetStatu", primaryjoin="AsaScience.science_goal_ouss_id == ObsUnitSetStatu.status_entity_id"
    )


class AsaScienceKeyword(Base):
    __tablename__ = "asa_science_keyword"
    __table_args__ = {"schema": "ALMA"}

    keyword = Column(String(100), nullable=False)
    category_id = Column(ForeignKey("ALMA.asa_science_keyword_category.id"), nullable=False, index=True)
    id = Column(Numeric(asdecimal=False), primary_key=True)

    category = relationship("AsaScienceKeywordCategory")


class AsaScienceKeywordCategory(Base):
    __tablename__ = "asa_science_keyword_category"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(4, 0, asdecimal=False), primary_key=True)
    catg_name = Column(String(100), nullable=False)


class AsaScienceTemp(Base):
    __tablename__ = "asa_science_temp"
    __table_args__ = (
        Index("asa_tmp_idx3", "source_name", "member_ouss_id"),
        Index("asa_tmp_idx2", "source_name", "group_ouss_id"),
        {"schema": "ALMA"},
    )

    dataset_id = Column(String(128), primary_key=True)
    asdm_uid = Column(String(32))
    field_uid = Column(String(32))
    spectral_window_uid = Column(String(32))
    project_uid = Column(String(32))
    schedblock_uid = Column(String(32))
    execblock_uid = Column(String(32))
    ra = Column(NullType, nullable=False)
    dec = Column(NullType, nullable=False)
    ra_source = Column(NullType, nullable=False)
    dec_source = Column(NullType, nullable=False)
    cx = Column(NullType)
    cy = Column(NullType)
    cz = Column(NullType)
    az_min = Column(NullType, nullable=False)
    az_max = Column(NullType, nullable=False)
    elevation_min = Column(NullType, nullable=False)
    elevation_max = Column(NullType, nullable=False)
    wcs_cd1_1 = Column(NullType)
    wcs_cd1_2 = Column(NullType)
    wcs_cd2_1 = Column(NullType)
    wcs_cd2_2 = Column(NullType)
    wcs_crpix1 = Column(NullType)
    wcs_crpix2 = Column(NullType)
    wcs_crval1 = Column(NullType)
    wcs_crval2 = Column(NullType)
    spatial_scale_min = Column(NullType, nullable=False)
    spatial_scale_max = Column(NullType, nullable=False)
    gal_longitude = Column(NullType, nullable=False)
    gal_latitude = Column(NullType, nullable=False)
    ecliptic_longitude = Column(NullType, nullable=False)
    ecliptic_latitude = Column(NullType, nullable=False)
    footprint = Column(String(2048))
    fov = Column(NullType, nullable=False)
    bounding_box = Column(String(2048))
    source_name = Column(String(256))
    source_is_in_simbad = Column(String(1), nullable=False)
    source_class = Column(String(128))
    source_redshift = Column(NullType)
    source_alma_catalog_id = Column(Numeric(22, 0, asdecimal=False))
    start_date = Column(DateTime, nullable=False)
    end_date = Column(DateTime, nullable=False)
    int_time = Column(Numeric(10, 3), nullable=False)
    time_resolution = Column(Numeric(8, 4), nullable=False)
    frequency = Column(NullType)
    band = Column(Numeric(scale=0, asdecimal=False))
    frequency_resolution = Column(NullType)
    velocity_resolution = Column(NullType)
    resolving_power = Column(NullType)
    bandwidth = Column(NullType)
    rest_frequency = Column(NullType)
    rest_frequency_min = Column(NullType)
    rest_frequency_max = Column(NullType)
    rest_frequency_resolution = Column(NullType)
    rest_velocity_resolution = Column(NullType, nullable=False)
    rest_resolving_power = Column(NullType, nullable=False)
    rest_frequency_support = Column(String(512), nullable=False)
    channel_num = Column(Numeric(scale=0, asdecimal=False))
    spectral_window_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    continuum_window_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    obs_unitset_id = Column(String(32), nullable=False)
    science_goal_ouss_id = Column(String(32))
    group_ouss_id = Column(String(32))
    member_ouss_id = Column(String(32))
    sensitivity = Column(Numeric(22, 0, asdecimal=False))
    flux_min = Column(NullType)
    flux_max = Column(NullType)
    uv_coverage_param1 = Column(NullType)
    uv_coverage_param2 = Column(NullType)
    pol_num = Column(Numeric(scale=0, asdecimal=False))
    pol_products = Column(String(64))
    airmass = Column(NullType)
    pwv = Column(NullType)
    temperature = Column(NullType)
    windspeed = Column(NullType)
    winddirection = Column(NullType)
    antennas = Column(String(660))
    baseline_max = Column(NullType, nullable=False)
    baseline_min = Column(NullType, nullable=False)
    scan_num = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    project_code = Column(String(20), nullable=False)
    schedblock_name = Column(String(128))
    observation_category = Column(String(128), nullable=False)
    scan_intent = Column(String(256))
    access_format = Column(String(10), nullable=False)
    processing_level = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    frequency_support = Column(String(4000))
    spatial_resolution = Column(NullType, nullable=False)
    spatres = Column(NullType)
    mineltp = Column(NullType)
    minel12 = Column(NullType)
    minel7 = Column(NullType)
    max_angscale = Column(NullType)
    obs_mode = Column(String(80))
    pipver = Column(String(80))
    asdm_list = Column(String(200))
    ant_tp_num = Column(Numeric(3, 0, asdecimal=False), nullable=False)
    ant_num = Column(Numeric(3, 0, asdecimal=False), nullable=False)
    ant_main_num = Column(Numeric(3, 0, asdecimal=False), nullable=False)
    ant_aca_num = Column(Numeric(3, 0, asdecimal=False), nullable=False)
    frequency_max = Column(NullType, nullable=False)
    frequency_min = Column(NullType, nullable=False)
    requestedarray = Column(String(8))
    asa_ous_id = Column(String(64))
    band_list = Column(String(30))
    antenna_string = Column(String(30))
    product_type = Column(String(10))
    dataset_product_id = Column(String(128))
    rms = Column(NullType)
    is_mosaic = Column(String(1), nullable=False)
    best_source_redshift = Column(NullType)
    redshift_provenance = Column(String(1))
    cont_sensitivity_channel = Column(NullType)
    cont_sensitivity_10kms = Column(NullType)
    cont_sensitivity_bandwidth = Column(NullType)
    parent_dataset_id = Column(String(128))


class AsaTable(Base):
    __tablename__ = "asa_tables"
    __table_args__ = (Index("asa_tables_unique", "schema_name", "table_name", unique=True), {"schema": "ALMA"})

    schema_name = Column(String(128), nullable=False)
    table_name = Column(String(128), nullable=False)
    table_type = Column(String(5), nullable=False)
    description = Column(String(1024))
    utype = Column(String(512))
    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)


class AsaTablesKey(Base):
    __tablename__ = "asa_tables_keys"
    __table_args__ = {"schema": "ALMA"}

    key_id = Column(String(128), primary_key=True)
    from_table = Column(String(128), nullable=False)
    target_table = Column(String(128), nullable=False)
    description = Column(String(1024), nullable=False)
    utype = Column(String(512))


class AsaTablesKeysColumn(Base):
    __tablename__ = "asa_tables_keys_columns"
    __table_args__ = {"schema": "ALMA"}

    key_id = Column(String(128), primary_key=True)
    from_column = Column(String(128), nullable=False)
    target_column = Column(String(128), nullable=False)


class AsaTapSchema(Base):
    __tablename__ = "asa_tap_schemas"
    __table_args__ = {"schema": "ALMA"}

    schema_name = Column(String(128), nullable=False, unique=True)
    description = Column(String(1024))
    utype = Column(String(512))
    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)


t_asp_accounts = Table(
    "asp_accounts",
    metadata,
    Column("account_id", String(32), nullable=False),
    Column("provides_user_demographics", String(1)),
    Column("last_demographics_page_visit", DateTime),
    schema="ALMA",
)


t_asp_obs_proposal = Table(
    "asp_obs_proposal",
    metadata,
    Column("archive_uid", String(33), nullable=False),
    Column("cycle", String(32)),
    Column("code", String(64), nullable=False),
    Column("priority", String(2)),
    Column("title", String(256)),
    Column("pi_id", String(32)),
    Column("affiliation", String(64)),
    Column("executive", String(32)),
    Column("qa0passcount", Numeric(asdecimal=False)),
    Column("final_scientific_category", String(32)),
    Column("project_state", String(32)),
    schema="ALMA",
)


t_asp_obs_proposal_author = Table(
    "asp_obs_proposal_author",
    metadata,
    Column("proposal_archive_uid", String(255), nullable=False),
    Column("author_index", Numeric(scale=0, asdecimal=False), nullable=False),
    Column("account_id", String(32), nullable=False),
    Column("firstname", String(256), nullable=False),
    Column("initials", String(256)),
    Column("lastname", String(256), nullable=False),
    schema="ALMA",
)


t_assignment = Table(
    "assignment",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("participant_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("score", Float),
    Column("proposal_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("state_change_date", DateTime),
    Column("status", String(32)),
    Column("assignment_type", String(32)),
    Column("conflict", String(32)),
    Column("normalized_score", Float),
    Column("vote_score", Float),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("rejection_type", String(32)),
    Column("normalized_vote_score", Numeric(asdecimal=False)),
    Column("assignment_comment", String(4000)),
    Column("reason_for_rejection", String(4000)),
    Column("comment_to_sa", String(4000)),
    Column("comment_to_phase2", String(4000)),
    Column("resurrection_justification", String(4000)),
    schema="ALMA",
)


t_assignment_view = Table(
    "assignment_view",
    metadata,
    Column("archive_uid", String(33), nullable=False),
    Column("proposal_code", String(64)),
    Column("cycle_code", String(32)),
    Column("title", String(256)),
    Column("pi_userid", String(32)),
    Column("abstract_text", Text),
    Column("scientific_category", String(32)),
    Column("new_scientific_category", String(16)),
    Column("overridden_sci_cat", String(32)),
    Column("proposal_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("feasible", String(8)),
    Column("arp_avg_score", Float),
    Column("arp_std_dev_scores", Float),
    Column("arp_score", Float),
    Column("arp_prop_strength", String(4000)),
    Column("arp_prop_weakness", String(4000)),
    Column("arp_prop_improvement", String(4000)),
    Column("arp_prop_evaluation", String(4000)),
    Column("arp_comment", String(4000)),
    Column("aprc_rank", Numeric(scale=0, asdecimal=False)),
    Column("aprc_letter_grade", String(1)),
    Column("aprc_prop_strength", String(4000)),
    Column("aprc_prop_weakness", String(4000)),
    Column("aprc_prop_improvement", String(4000)),
    Column("aprc_prop_evaluation", String(4000)),
    Column("aprc_comment", String(4000)),
    Column("dc_letter_grade", String(1)),
    Column("assignment_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("status", String(32)),
    Column("conflict", String(32)),
    Column("assignment_type", String(32)),
    Column("assignment_comment", String(4000)),
    Column("comment_to_sa", String(4000)),
    Column("comment_to_phase2", String(4000)),
    Column("score", Float),
    Column("normalized_score", Float),
    Column("reason_for_rejection", String(4000)),
    Column("assessor_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("participant_type", String(64), nullable=False),
    Column("assessor_userid", String(32), nullable=False),
    Column("assessor_first_name", String(256)),
    Column("assessor_last_name", String(256)),
    Column("assessor_email", String(256)),
    Column("arp_id", Numeric(38, 0, asdecimal=False)),
    Column("panel_name", String(255)),
    Column("role_on_panel", String(255)),
    schema="ALMA",
)


class BmmvObsproject(Base):
    __tablename__ = "bmmv_obsproject"
    __table_args__ = {"schema": "ALMA"}

    prj_archive_uid = Column(String(33), primary_key=True)
    deleted = Column(Numeric(1, 0, asdecimal=False), index=True)
    pi = Column(String(64), index=True)
    prj_name = Column(String(256))
    title = Column(String(256))
    array = Column(String(64))
    prj_code = Column(String(64), nullable=False)
    code = Column(String(64), index=True)
    prj_time_of_creation = Column(String(23))
    prj_scientific_rank = Column(Numeric(8, 0, asdecimal=False))
    prj_version = Column(String(32))
    prj_assigned_priority = Column(Numeric(asdecimal=False))
    prj_letter_grade = Column(String(2))
    p2g_attention = Column(String(5), index=True)
    p2g_attention_reasons = Column(String(4000))
    manual_mode = Column(String(5))


class BmmvObsproposal(Base):
    __tablename__ = "bmmv_obsproposal"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    deleted = Column(Numeric(1, 0, asdecimal=False))
    abstract_text = Column(Text)
    scientific_category = Column(String(32))
    proposal_type = Column(String(32))
    pi_userid = Column(String(32))
    associatedexec = Column(String(32))
    datereceived = Column(String(32))
    obsproject_archive_uid = Column(String(32), nullable=False, index=True)
    projectuid = Column(String(32), nullable=False, index=True)
    pi_fullname = Column(String(64))
    organization = Column(String(64))
    email = Column(String(64))
    cycle = Column(String(32), index=True)
    keyword1 = Column(String(128))
    keyword2 = Column(String(128))
    keywordcode1 = Column(String(128))
    keywordcode2 = Column(String(128))
    studentproject = Column(String(32))


class BmmvObsproposalAuthor(Base):
    __tablename__ = "bmmv_obsproposal_authors"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), nullable=False, index=True)
    deleted = Column(Numeric(1, 0, asdecimal=False))
    cycle = Column(String(32), nullable=False)
    project_archive_uid = Column(String(33), nullable=False)
    userid = Column(String(32), nullable=False)
    organisation = Column(String(32))
    executive = Column(String(32))
    authtype = Column(String(4), nullable=False)
    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    organisation_id = Column(Numeric(38, 0, asdecimal=False))


class BmmvObsproposalFeedback(Base):
    __tablename__ = "bmmv_obsproposal_feedback"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    nonalmaexectimefraction = Column(String(16))
    chileexectimefraction = Column(String(16))
    eaexectimefraction = Column(String(16))
    euexectimefraction = Column(String(16))
    naexectimefraction = Column(String(16))
    integration_time = Column(String(64))
    integration_time_unit = Column(String(64))
    receiverband01time = Column(Numeric(asdecimal=False))
    receiverband02time = Column(Numeric(asdecimal=False))
    receiverband03time = Column(Numeric(asdecimal=False))
    receiverband04time = Column(Numeric(asdecimal=False))
    receiverband05time = Column(Numeric(asdecimal=False))
    receiverband06time = Column(Numeric(asdecimal=False))
    receiverband07time = Column(Numeric(asdecimal=False))
    receiverband08time = Column(Numeric(asdecimal=False))
    receiverband09time = Column(Numeric(asdecimal=False))
    receiverband10time = Column(Numeric(asdecimal=False))
    receiverbanduntime = Column(Numeric(asdecimal=False))
    array_time_alma = Column(Numeric(asdecimal=False))
    array_time_unit_alma = Column(String(8))
    array_time_aca = Column(Numeric(asdecimal=False))
    array_time_unit_aca = Column(String(8))
    array_time_twelve_m = Column(Numeric(asdecimal=False))
    array_time_unit_twelve_m = Column(String(8))
    array_time_seven_m = Column(Numeric(asdecimal=False))
    array_time_unit_seven_m = Column(String(8))
    array_time_tp_array = Column(Numeric(asdecimal=False))
    array_time_unit_tp_array = Column(String(8))
    max_data_rate_twelve_m = Column(Numeric(asdecimal=False))
    max_data_rate_unit_twelve_m = Column(String(8))
    data_volume_twelve_m = Column(Numeric(asdecimal=False))
    data_volume_unit_twelve_m = Column(String(8))
    max_data_rate_seven_m = Column(Numeric(asdecimal=False))
    max_data_rate_unit_seven_m = Column(String(8))
    data_volume_seven_m = Column(Numeric(asdecimal=False))
    data_volume_unit_seven_m = Column(String(8))
    max_data_rate_tp_array = Column(Numeric(asdecimal=False))
    max_data_rate_unit_tp_array = Column(String(8))
    data_volume_tp_array = Column(Numeric(asdecimal=False))
    data_volume_unit_tp_array = Column(String(8))


class BmmvObsproposalTarget(Base):
    __tablename__ = "bmmv_obsproposal_target"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    prop_archive_uid = Column(String(33), nullable=False, index=True)
    science_goal_name = Column(String(1000))
    source_coord_system = Column(String(10))
    target_longitude = Column(Numeric(asdecimal=False))
    target_longitude_unit = Column(String(10))
    target_latitude = Column(Numeric(asdecimal=False))
    target_latitude_unit = Column(String(10))
    desired_sensitivity = Column(Numeric(asdecimal=False))
    desired_sensitivity_unit = Column(String(10))
    desired_angular_resolution = Column(Numeric(asdecimal=False))
    desired_ang_res_unit = Column(String(10))
    spectral_resolution = Column(Numeric(asdecimal=False))
    spectral_resolution_unit = Column(String(10))
    center_frequency = Column(Numeric(asdecimal=False))
    center_frequency_unit = Column(String(10))
    band_width = Column(Numeric(asdecimal=False))
    band_width_unit = Column(String(10))
    is_sky_frequency = Column(String(5))
    group_index = Column(Numeric(asdecimal=False))
    science_goal_index = Column(Numeric(asdecimal=False))
    start_frequency = Column(Numeric(asdecimal=False))
    start_frequency_unit = Column(String(10))
    end_frequency = Column(Numeric(asdecimal=False))
    end_frequency_unit = Column(String(10))


class BmmvObsunitset(Base):
    __tablename__ = "bmmv_obsunitset"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    archive_uid = Column(String(32), nullable=False, index=True)
    partid = Column(String(64))
    name = Column(String(256))
    scienceprocessingscript = Column(String(1024))
    runsciencepipeline = Column(String(32))
    numberschedblocks = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    ousstatusref = Column(String(32))
    requestedarray = Column(String(8))
    is_resubmission = Column(String(5))
    resubmission_of_name = Column(String(1000))
    resolution_option = Column(String(32))
    solar_system_object = Column(String(32))
    sg_mode = Column(String(16))


class BmmvProjTimeConstraint(Base):
    __tablename__ = "bmmv_proj_time_constraint"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    prj_archive_uid = Column(String(33), nullable=False, index=True)
    deleted = Column(Numeric(1, 0, asdecimal=False))
    time_constraint_type = Column(String(32))
    start_time = Column(String(32))
    end_time = Column(String(32))
    allowed_margin = Column(String(32))
    allowed_margin_unit = Column(String(5))
    required_delay = Column(String(32))
    required_delay_unit = Column(String(5))
    visit_id = Column(String(32))
    previous_visit_id = Column(String(32))
    monitoring_length = Column(String(32))
    monitoring_length_unit = Column(String(5))


class BmmvQuicklook(Base):
    __tablename__ = "bmmv_quicklook"
    __table_args__ = (
        Index("bmmv_quicklook_uni", "ousstatus_uid", "ousstatus_part_id", unique=True),
        {"schema": "ALMA"},
    )

    archive_uid = Column(String(33), primary_key=True)
    schedblock_uid = Column(String(64))
    qlfocussummary_uid = Column(String(64))
    qlpointingsummary_uid = Column(String(64))
    qlatmospheresummary_uid = Column(String(64))
    ousstatus_uid = Column(String(64), nullable=False)
    ousstatus_part_id = Column(String(64), nullable=False)
    qlphasesummary_uid = Column(String(64))


class BmmvSchedblock(Base):
    __tablename__ = "bmmv_schedblock"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    prj_ref = Column(String(32), index=True)
    receiver_band = Column(String(32))
    frequency = Column(String(32))
    ref_ra = Column(String(32))
    ref_dec = Column(String(32))
    sb_name = Column(String(32))
    time_string = Column(String(32))
    time_units = Column(String(32))
    receiverband = Column(String(32))
    refra = Column(String(32))
    refdec = Column(String(32))
    sbname = Column(String(32))
    totalestimatedasstring = Column(String(32))
    totalestimatedtimeunits = Column(String(32))
    status = Column(String(32))
    totalestimatedtime = Column(String(32))
    execution_count = Column(String(32))
    requestedarray = Column(String(8))
    minacceptableangresolution = Column(Float)
    maxacceptableangresolution = Column(Float)
    representativefrequency = Column(Float)
    nominalconfiguration = Column(String(32))
    gous_status_uid = Column(String(32))
    mous_status_uid = Column(String(32))
    sgous_status_uid = Column(String(32))


class BmmvSchedblockPolarisation(Base):
    __tablename__ = "bmmv_schedblock_polarisation"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    archive_uid = Column(String(33), nullable=False, index=True)
    products = Column(String(32), nullable=False)


class BmmvSchedblockSource(Base):
    __tablename__ = "bmmv_schedblock_sources"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    archive_uid = Column(String(33), nullable=False, index=True)
    field_source_name = Column(String(32), nullable=False)


class Country(Base):
    __tablename__ = "country"
    __table_args__ = {"schema": "ALMA"}

    country_id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    country_name = Column(String(256), nullable=False, unique=True)
    executive = Column(String(32), nullable=False)


t_cycle = Table(
    "cycle",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("code", String(255), nullable=False),
    Column("cycle_step", String(64), nullable=False),
    Column("total_hours", Numeric(8, 0, asdecimal=False)),
    Column("ea_percent", Numeric(8, 2)),
    Column("eu_percent", Numeric(8, 2)),
    Column("na_percent", Numeric(8, 2)),
    Column("cl_percent", Numeric(8, 2)),
    Column("other_percent", Numeric(8, 2)),
    Column("percent_marker_1", Numeric(8, 0, asdecimal=False)),
    Column("percent_marker_2", Numeric(8, 0, asdecimal=False)),
    Column("percent_marker_3", Numeric(8, 0, asdecimal=False)),
    Column("scientific_categories", String(255)),
    Column("start_date", DateTime),
    Column("end_date", DateTime),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("triage_std_dev_factor", Numeric(8, 2)),
    Column("triage_percent", Numeric(8, 2)),
    Column("ddt", String(1)),
    Column("active", String(1)),
    Column("eu_triage_percent", Numeric(8, 2)),
    Column("na_triage_percent", Numeric(8, 2)),
    Column("ea_triage_percent", Numeric(8, 2)),
    Column("cl_triage_percent", Numeric(8, 2)),
    Column("duplication_arc_seconds", Numeric(8, 2)),
    Column("keyword_codes", String(4000)),
    Column("score_mean", Numeric(8, 6)),
    Column("score_std_dev", Numeric(8, 6)),
    schema="ALMA",
)


t_dbmaintain_scripts = Table(
    "dbmaintain_scripts",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


t_dbmaintain_scripts_account = Table(
    "dbmaintain_scripts_account",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


t_dbmaintain_scripts_apdm = Table(
    "dbmaintain_scripts_apdm",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


t_dbmaintain_scripts_asa = Table(
    "dbmaintain_scripts_asa",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


t_dbmaintain_scripts_ph1m = Table(
    "dbmaintain_scripts_ph1m",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


t_dbmaintain_scripts_protrack = Table(
    "dbmaintain_scripts_protrack",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


t_dbmaintain_scripts_version = Table(
    "dbmaintain_scripts_version",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


class DpCacheAsdmInfo(Base):
    __tablename__ = "dp_cache_asdm_info"
    __table_args__ = {"schema": "ALMA"}

    asdm_id = Column(String(64), primary_key=True)
    tar_size_bytes = Column(Numeric(12, 0, asdecimal=False))


class DpCacheXmlInfo(Base):
    __tablename__ = "dp_cache_xml_info"
    __table_args__ = {"schema": "ALMA"}

    xml_id = Column(String(64), primary_key=True)
    size_bytes = Column(Numeric(9, 0, asdecimal=False))


t_dp_delegation = Table(
    "dp_delegation",
    metadata,
    Column("id", Numeric(22, 0, asdecimal=False), nullable=False),
    Column("pi_rh_id", Numeric(22, 0, asdecimal=False), nullable=False),
    Column("project_code", String(18), nullable=False),
    Column("delegee_rh_id", Numeric(22, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


t_dp_delegation_backup = Table(
    "dp_delegation_backup",
    metadata,
    Column("id", Numeric(22, 0, asdecimal=False), nullable=False),
    Column("pi_rh_id", Numeric(22, 0, asdecimal=False), nullable=False),
    Column("project_code", String(18), nullable=False),
    Column("delegee_rh_id", Numeric(22, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


t_dp_delivery_asdm_ous_bkp = Table(
    "dp_delivery_asdm_ous_bkp",
    metadata,
    Column("asdm_uid", String(33)),
    Column("ous_status_id", String(33)),
    Column("deliverable_name", String(64), nullable=False),
    Column("id", Numeric(22, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


t_drm_data_reducer = Table(
    "drm_data_reducer",
    metadata,
    Column("dr_user_id", ForeignKey("ALMA.account.account_id"), primary_key=True),
    Column("node_id", ForeignKey("ALMA.drm_node.id")),
    schema="ALMA",
)


class DrmDrHistory(Base):
    __tablename__ = "drm_dr_history"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    author = Column(String(32))
    entityuid = Column(String(32), nullable=False)
    datareducer = Column(String(32))
    ttimestamp = Column(DateTime, nullable=False)
    arc = Column(String(8))


class DrmDrQualification(Base):
    __tablename__ = "drm_dr_qualification"
    __table_args__ = {"schema": "ALMA"}

    dr_user_id = Column(ForeignKey("ALMA.account.account_id"), primary_key=True, nullable=False)
    qualification = Column(String(32), primary_key=True, nullable=False)

    dr_user = relationship("Account")


class DrmNode(Base):
    __tablename__ = "drm_node"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(3, 0, asdecimal=False), primary_key=True)
    arc = Column(String(8), nullable=False)
    node = Column(String(32))


t_email_template = Table(
    "email_template",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("name", String(255), nullable=False),
    Column("content_type", String(64)),
    Column("subject", String(255)),
    Column("body", Text),
    Column("query", Text),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("system_email", String(1)),
    schema="ALMA",
)


class ErrorLog(Base):
    __tablename__ = "error_log"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(22, 0, asdecimal=False), primary_key=True)
    archive_uid = Column(String(32))
    table_name = Column(String(30))
    ora_error_code = Column(Numeric(scale=0, asdecimal=False))
    ora_error_message = Column(String(4000))
    ora_backtrace = Column(Text)
    ora_callstack = Column(Text)
    created_on = Column(DateTime)
    created_by = Column(String(30))


class Eu1(Base):
    __tablename__ = "eu1"
    __table_args__ = {"schema": "ALMA"}

    data1 = Column(String(100), primary_key=True)


class Institution(Base):
    __tablename__ = "institution"
    __table_args__ = (
        Index("institution_name_unique", "name1", "name2", "name3", "country_id", "state", unique=True),
        {"schema": "ALMA"},
    )

    inst_no = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    sibling_no = Column(ForeignKey("ALMA.institution.inst_no"))
    name1 = Column(String(2048), nullable=False)
    name2 = Column(String(256))
    name3 = Column(String(256))
    altnames = Column(String(256))
    city = Column(String(64))
    postcode = Column(String(32))
    executive = Column(String(5), nullable=False)
    state = Column(String(32))
    email = Column(String(256))
    url = Column(String(256))
    phone = Column(String(128))
    fax = Column(String(128))
    recorder = Column(String(256))
    superseding_no = Column(Numeric(38, 0, asdecimal=False))
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False, server_default=text("0 "))
    country_id = Column(ForeignKey("ALMA.country.country_id"))
    user_defined = Column(String(1), server_default=text("'F'"))
    ignored = Column(String(1), server_default=text("'F'"))
    address1 = Column(String(512))
    address2 = Column(String(512))
    address3 = Column(String(512))

    country = relationship("Country")
    parent = relationship("Institution", remote_side=[inst_no])


t_mv_obsproject = Table(
    "mv_obsproject",
    metadata,
    Column("prj_archive_uid", String(33), nullable=False),
    Column("pi", String(32)),
    Column("prj_name", String(256)),
    Column("array", String(64)),
    Column("prj_code", String(64), nullable=False),
    Column("prj_time_of_creation", String(23)),
    Column("prj_scientific_rank", Numeric(8, 0, asdecimal=False)),
    Column("prj_version", String(32)),
    Column("prj_assigned_priority", Numeric(asdecimal=False)),
    Column("prj_letter_grade", String(2)),
    Column("ph1m_priority_grade", String(2)),
    Column("pi_fullname", String(64)),
    Column("organization", String(64)),
    Column("email", String(64)),
    Column("cycle", String(32)),
    Column("executive", String(32)),
    Column("p2g_attention", String(5)),
    Column("p2g_attention_reasons", String(4000)),
    schema="ALMA",
)


t_mv_obsproject_derived = Table(
    "mv_obsproject_derived",
    metadata,
    Column("prj_archive_uid", String(33), nullable=False),
    Column("total_time", Numeric(asdecimal=False)),
    Column("used_time", Numeric(asdecimal=False)),
    schema="ALMA",
)


t_mv_obsproposal = Table(
    "mv_obsproposal",
    metadata,
    Column("archive_uid", String(33), nullable=False),
    Column("deleted", Numeric(1, 0, asdecimal=False)),
    Column("title", String(256)),
    Column("code", String(64)),
    Column("cycle", String(32)),
    Column("abstract_text", Text),
    Column("scientific_category", String(32)),
    Column("proposal_type", String(32)),
    Column("pi_userid", String(32)),
    Column("obsproject_archive_uid", String(32), nullable=False),
    Column("pi_executive", String(32)),
    Column("datereceived", String(32)),
    Column("keyword1", String(128)),
    Column("keyword2", String(128)),
    Column("keywordcode1", String(128)),
    Column("keywordcode2", String(128)),
    Column("studentproject", String(32)),
    Column("nonalmaexectimefraction", String(16)),
    Column("chileexectimefraction", String(16)),
    Column("eaexectimefraction", String(16)),
    Column("euexectimefraction", String(16)),
    Column("naexectimefraction", String(16)),
    Column("integration_time", Numeric(asdecimal=False)),
    Column("integration_time_unit", String(8)),
    Column("receiverband01time", Numeric(asdecimal=False)),
    Column("receiverband02time", Numeric(asdecimal=False)),
    Column("receiverband03time", Numeric(asdecimal=False)),
    Column("receiverband04time", Numeric(asdecimal=False)),
    Column("receiverband05time", Numeric(asdecimal=False)),
    Column("receiverband06time", Numeric(asdecimal=False)),
    Column("receiverband07time", Numeric(asdecimal=False)),
    Column("receiverband08time", Numeric(asdecimal=False)),
    Column("receiverband09time", Numeric(asdecimal=False)),
    Column("receiverband10time", Numeric(asdecimal=False)),
    Column("receiverbanduntime", Numeric(asdecimal=False)),
    Column("array_time_alma", Numeric(asdecimal=False)),
    Column("array_time_unit_alma", String(8)),
    Column("array_time_twelve_m", Numeric(asdecimal=False)),
    Column("array_time_unit_twelve_m", String(8)),
    Column("array_time_aca", Numeric(asdecimal=False)),
    Column("array_time_unit_aca", String(8)),
    Column("array_time_seven_m", Numeric(asdecimal=False)),
    Column("array_time_unit_seven_m", String(8)),
    Column("array_time_tp_array", Numeric(asdecimal=False)),
    Column("array_time_unit_tp_array", String(8)),
    Column("max_data_rate_twelve_m", Numeric(asdecimal=False)),
    Column("max_data_rate_unit_twelve_m", String(8)),
    Column("data_volume_twelve_m", Numeric(asdecimal=False)),
    Column("data_volume_unit_twelve_m", String(8)),
    Column("max_data_rate_seven_m", Numeric(asdecimal=False)),
    Column("max_data_rate_unit_seven_m", String(8)),
    Column("data_volume_seven_m", Numeric(asdecimal=False)),
    Column("data_volume_unit_seven_m", String(8)),
    Column("max_data_rate_tp_array", Numeric(asdecimal=False)),
    Column("max_data_rate_unit_tp_array", String(8)),
    Column("data_volume_tp_array", Numeric(asdecimal=False)),
    Column("data_volume_unit_tp_array", String(8)),
    Column("domain_entity_state", String(32), nullable=False),
    Column("status_entity_id", String(64), nullable=False),
    schema="ALMA",
)


t_mv_obsproposal_authors = Table(
    "mv_obsproposal_authors",
    metadata,
    Column("archive_uid", String(33), nullable=False),
    Column("prj_code", String(64), nullable=False),
    Column("deleted", Numeric(1, 0, asdecimal=False)),
    Column("cycle", String(32), nullable=False),
    Column("project_archive_uid", String(33), nullable=False),
    Column("userid", String(32), nullable=False),
    Column("organisation", String(32)),
    Column("organisation_id", Numeric(38, 0, asdecimal=False)),
    Column("executive", String(32)),
    Column("authtype", String(4), nullable=False),
    schema="ALMA",
)


t_mv_obsproposal_target = Table(
    "mv_obsproposal_target",
    metadata,
    Column("id", Numeric(asdecimal=False), nullable=False),
    Column("prop_archive_uid", String(33), nullable=False),
    Column("cycle", String(32)),
    Column("code", String(64)),
    Column("scientific_category", String(32)),
    Column("new_scientific_category", String(16)),
    Column("science_goal_name", String(1000)),
    Column("source_coord_system", String(10)),
    Column("target_longitude", Numeric(asdecimal=False)),
    Column("target_longitude_unit", String(10)),
    Column("target_latitude", Numeric(asdecimal=False)),
    Column("target_latitude_unit", String(10)),
    Column("desired_sensitivity", Numeric(asdecimal=False)),
    Column("desired_sensitivity_unit", String(10)),
    Column("desired_angular_resolution", Numeric(asdecimal=False)),
    Column("desired_ang_res_unit", String(10)),
    Column("spectral_resolution", Numeric(asdecimal=False)),
    Column("spectral_resolution_unit", String(10)),
    Column("center_frequency", Numeric(asdecimal=False)),
    Column("center_frequency_unit", String(10)),
    Column("band_width", Numeric(asdecimal=False)),
    Column("band_width_unit", String(10)),
    Column("is_sky_frequency", String(4)),
    Column("group_index", Numeric(asdecimal=False)),
    Column("aprc_letter_grade", String(1)),
    Column("dc_letter_grade", String(1)),
    Column("science_goal_index", Numeric(asdecimal=False)),
    Column("project_status", String(32), nullable=False),
    Column("start_frequency", Numeric(asdecimal=False)),
    Column("start_frequency_unit", String(10)),
    Column("end_frequency", Numeric(asdecimal=False)),
    Column("end_frequency_unit", String(10)),
    schema="ALMA",
)


t_mv_obsunitset = Table(
    "mv_obsunitset",
    metadata,
    Column("archive_uid", String(32), nullable=False),
    Column("partid", String(64)),
    Column("name", String(256)),
    Column("scienceprocessingscript", String(1024)),
    Column("runsciencepipeline", String(32)),
    Column("numberschedblocks", String(32)),
    Column("requestedarray", String(8)),
    Column("ousstatusref", String(32)),
    Column("is_resubmission", String(5)),
    Column("resubmission_of_name", String(1000)),
    Column("resolution_option", String(32)),
    Column("solar_system_object", String(32)),
    Column("sg_mode", String(16)),
    schema="ALMA",
)


t_mv_proj_time_constraint = Table(
    "mv_proj_time_constraint",
    metadata,
    Column("id", Numeric(asdecimal=False), nullable=False),
    Column("prj_archive_uid", String(33), nullable=False),
    Column("deleted", Numeric(1, 0, asdecimal=False)),
    Column("time_constraint_type", String(32)),
    Column("start_time", String(32)),
    Column("end_time", String(32)),
    Column("allowed_margin", String(32)),
    Column("allowed_margin_unit", String(5)),
    Column("required_delay", String(32)),
    Column("required_delay_unit", String(5)),
    Column("visit_id", String(32)),
    Column("previous_visit_id", String(32)),
    Column("monitoring_length", String(32)),
    Column("monitoring_length_unit", String(5)),
    schema="ALMA",
)


t_mv_quicklook = Table(
    "mv_quicklook",
    metadata,
    Column("archive_uid", String(33), nullable=False),
    Column("schedblock_uid", String(64)),
    Column("qlfocussummary_uid", String(64)),
    Column("qlpointingsummary_uid", String(64)),
    Column("qlatmospheresummary_uid", String(64)),
    schema="ALMA",
)


t_mv_schedblock = Table(
    "mv_schedblock",
    metadata,
    Column("sb_archive_uid", String(33), nullable=False),
    Column("prj_ref", String(32)),
    Column("receiver_band", String(32)),
    Column("frequency", String(32)),
    Column("ref_ra", String(32)),
    Column("ref_dec", String(32)),
    Column("sb_name", String(32)),
    Column("requested_array", String(8)),
    Column("minacceptableangresolution", Float),
    Column("maxacceptableangresolution", Float),
    Column("representativefrequency", Float),
    Column("status", String(32), nullable=False),
    Column("used_time", Numeric(asdecimal=False)),
    Column("time_string", String(32)),
    Column("time_units", String(32)),
    Column("execution_count", String(32)),
    Column("nominalconfiguration", String(32)),
    Column("time", Numeric(asdecimal=False)),
    Column("phase1_flag", Numeric(asdecimal=False)),
    schema="ALMA",
)


t_mv_schedblock_sourcename = Table(
    "mv_schedblock_sourcename",
    metadata,
    Column("id", Numeric(asdecimal=False), nullable=False),
    Column("archive_uid", String(33), nullable=False),
    Column("sourcename", String(32), nullable=False),
    schema="ALMA",
)


class ObsProjectStatu(Base):
    __tablename__ = "obs_project_status"
    __table_args__ = {"schema": "ALMA"}

    status_entity_id = Column(String(64), primary_key=True)
    domain_entity_id = Column(String(64), nullable=False, index=True)
    domain_entity_state = Column(String(32), nullable=False, index=True)
    obs_project_status_id = Column(String(64), nullable=False)
    obs_program_status_id = Column(String(64), nullable=False)
    obs_project_id = Column(ForeignKey("ALMA.xml_obsproject_entities.archive_uid"), nullable=False, unique=True)
    project_was_timed_out = Column(DateTime)
    xml = Column(NullType)
    flags = Column(String(200), index=True)

    obs_project = relationship("XmlObsprojectEntity")


class ObsUnitSetStatu(Base):
    __tablename__ = "obs_unit_set_status"
    __table_args__ = (Index("ous_status_project_entity_idx", "obs_project_id", "domain_entity_id"), {"schema": "ALMA"})

    status_entity_id = Column(String(64), primary_key=True)
    domain_entity_id = Column(String(64))
    domain_entity_state = Column(String(32), nullable=False)
    parent_obs_unit_set_status_id = Column(String(64))
    obs_project_status_id = Column(String(64), nullable=False)
    obs_project_id = Column(ForeignKey("ALMA.xml_obsproject_entities.archive_uid"), nullable=False)
    total_required_time_in_sec = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    total_used_time_in_sec = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    xml = Column(NullType)
    flags = Column(String(200), index=True)

    obs_project = relationship("XmlObsprojectEntity")


class OcdObservingCycle(Base):
    __tablename__ = "ocd_observing_cycle"
    __table_args__ = {"schema": "ALMA"}

    observingcycleid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    cycle = Column(String(32), nullable=False, unique=True)
    description = Column(String(256))
    clfraction = Column(Numeric(8, 2))
    eafraction = Column(Numeric(8, 2))
    eufraction = Column(Numeric(8, 2))
    nafraction = Column(Numeric(8, 2))
    otherfraction = Column(Numeric(8, 2))


class OcdObservingSession(Base):
    __tablename__ = "ocd_observing_session"
    __table_args__ = (
        Index("ocd_obssession_uni", "observingcycleid", "blocknumber", "startdate", unique=True),
        {"schema": "ALMA"},
    )

    sessionid = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    observingcycleid = Column(ForeignKey("ALMA.ocd_observing_cycle.observingcycleid"), nullable=False)
    blocknumber = Column(Numeric(8, 0, asdecimal=False), nullable=False)
    startdate = Column(DateTime, nullable=False)
    enddate = Column(DateTime, nullable=False)
    comments = Column(String(256))

    ocd_observing_cycle = relationship("OcdObservingCycle")


t_old_dbmaintain_scripts = Table(
    "old_dbmaintain_scripts",
    metadata,
    Column("file_name", String(150)),
    Column("file_last_modified_at", Numeric(scale=0, asdecimal=False)),
    Column("checksum", String(50)),
    Column("executed_at", String(20)),
    Column("succeeded", Numeric(scale=0, asdecimal=False)),
    schema="ALMA",
)


class OusOperation(Base):
    __tablename__ = "ous_operations"
    __table_args__ = {"schema": "ALMA"}

    obs_unit_set_id = Column(String(64), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    data_reducer_account_id = Column(ForeignKey("ALMA.account.account_id"))
    data_reducer_executive = Column(String(64))
    qa2recipients = Column(String(512))
    drm_account_id = Column(ForeignKey("ALMA.account.account_id"))

    data_reducer_account = relationship(
        "Account", primaryjoin="OusOperation.data_reducer_account_id == Account.account_id"
    )
    drm_account = relationship("Account", primaryjoin="OusOperation.drm_account_id == Account.account_id")


t_panel = Table(
    "panel",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("cycle_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("panel_name", String(255), nullable=False),
    Column("minutes", Text),
    Column("panel_type", String(32), nullable=False),
    Column("scientific_category", String(32)),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    schema="ALMA",
)


t_panel_member = Table(
    "panel_member",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("participant_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("panel_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("role_on_panel", String(255)),
    Column("package_downloaded", DateTime),
    schema="ALMA",
)


t_participant = Table(
    "participant",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("cycle_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("package_downloaded", DateTime),
    Column("userid", String(32), nullable=False),
    Column("participant_type", String(64), nullable=False),
    Column("alma_location", String(32)),
    Column("keyword_codes", String(256)),
    Column("score_mean", Numeric(8, 6)),
    Column("score_std_dev", Numeric(8, 6)),
    schema="ALMA",
)


class Ph1mAssignment(Base):
    __tablename__ = "ph1m_assignment"
    __table_args__ = (Index("assignment_unique", "proposal_id", "participant_id", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    participant_id = Column(ForeignKey("ALMA.ph1m_participant.id"), nullable=False, index=True)
    score = Column(Float)
    proposal_id = Column(ForeignKey("ALMA.ph1m_proposal.id"), nullable=False, index=True)
    state_change_date = Column(DateTime)
    status = Column(String(32))
    assignment_type = Column(String(32))
    conflict = Column(String(32))
    normalized_score = Column(Float)
    vote_score = Column(Float)
    update_date = Column(DateTime)
    update_userid = Column(String(32))
    rejection_type = Column(String(32))
    normalized_vote_score = Column(Numeric(asdecimal=False))
    assignment_comment = Column(String(4000))
    reason_for_rejection = Column(String(4000))
    comment_to_sa = Column(String(4000))
    comment_to_phase2 = Column(String(4000))
    resurrection_justification = Column(String(4000))

    participant = relationship("Ph1mParticipant")
    proposal = relationship("Ph1mProposal")


class Ph1mAutosave(Base):
    __tablename__ = "ph1m_autosave"
    __table_args__ = {"schema": "ALMA"}

    table_name = Column(String(32), primary_key=True, nullable=False)
    column_name = Column(String(32), primary_key=True, nullable=False)
    row_id = Column(Numeric(38, 0, asdecimal=False), primary_key=True, nullable=False)
    user_id = Column(String(32), primary_key=True, nullable=False)
    update_date = Column(DateTime, nullable=False)
    saved_text = Column(Text, nullable=False)


class Ph1mCycle(Base):
    __tablename__ = "ph1m_cycle"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    code = Column(String(255), nullable=False, unique=True)
    cycle_step = Column(String(64), nullable=False)
    total_hours = Column(Numeric(8, 0, asdecimal=False))
    ea_percent = Column(Numeric(8, 2))
    eu_percent = Column(Numeric(8, 2))
    na_percent = Column(Numeric(8, 2))
    cl_percent = Column(Numeric(8, 2))
    other_percent = Column(Numeric(8, 2))
    percent_marker_1 = Column(Numeric(8, 0, asdecimal=False))
    percent_marker_2 = Column(Numeric(8, 0, asdecimal=False))
    percent_marker_3 = Column(Numeric(8, 0, asdecimal=False))
    scientific_categories = Column(String(255))
    start_date = Column(DateTime)
    end_date = Column(DateTime)
    update_date = Column(DateTime)
    update_userid = Column(String(32))
    triage_std_dev_factor = Column(Numeric(8, 2))
    triage_percent = Column(Numeric(8, 2))
    ddt = Column(String(1), server_default=text("'F'"))
    active = Column(String(1), server_default=text("'T'"))
    eu_triage_percent = Column(Numeric(8, 2), server_default=text("70"))
    na_triage_percent = Column(Numeric(8, 2), server_default=text("70"))
    ea_triage_percent = Column(Numeric(8, 2), server_default=text("70"))
    cl_triage_percent = Column(Numeric(8, 2), server_default=text("70"))
    duplication_arc_seconds = Column(Numeric(8, 2), server_default=text("120"))
    keyword_codes = Column(String(4000))
    score_mean = Column(Numeric(8, 6))
    score_std_dev = Column(Numeric(8, 6))
    triage_oversub_percent = Column(
        Numeric(8, 2),
        server_default=text(
            """\
300
   """
        ),
    )


class Ph1mEmailTemplate(Base):
    __tablename__ = "ph1m_email_template"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    name = Column(String(255), nullable=False, unique=True)
    content_type = Column(String(64))
    subject = Column(String(255))
    body = Column(Text)
    query = Column(Text)
    update_date = Column(DateTime)
    update_userid = Column(String(32))
    system_email = Column(
        String(1),
        server_default=text(
            """\
'F'
   """
        ),
    )


class Ph1mPanel(Base):
    __tablename__ = "ph1m_panel"
    __table_args__ = (Index("unique_panel_name", "cycle_id", "panel_name", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    cycle_id = Column(ForeignKey("ALMA.ph1m_cycle.id"), nullable=False)
    panel_name = Column(String(255), nullable=False)
    minutes = Column(Text)
    panel_type = Column(String(32), nullable=False)
    scientific_category = Column(String(32))
    update_date = Column(DateTime)
    update_userid = Column(String(32))

    cycle = relationship("Ph1mCycle")
    proposals = relationship("Ph1mProposal", secondary="ALMA.ph1m_panel_proposal")


class Ph1mPanelMember(Base):
    __tablename__ = "ph1m_panel_member"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    participant_id = Column(ForeignKey("ALMA.ph1m_participant.id"), nullable=False)
    panel_id = Column(ForeignKey("ALMA.ph1m_panel.id"), nullable=False)
    role_on_panel = Column(String(255))
    package_downloaded = Column(DateTime)

    panel = relationship("Ph1mPanel")
    participant = relationship("Ph1mParticipant")


t_ph1m_panel_proposal = Table(
    "ph1m_panel_proposal",
    metadata,
    Column("panel_id", ForeignKey("ALMA.ph1m_panel.id"), primary_key=True, nullable=False),
    Column("proposal_id", ForeignKey("ALMA.ph1m_proposal.id"), primary_key=True, nullable=False),
    schema="ALMA",
)


class Ph1mParticipant(Base):
    __tablename__ = "ph1m_participant"
    __table_args__ = (Index("participant_unique", "userid", "cycle_id", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    cycle_id = Column(ForeignKey("ALMA.ph1m_cycle.id"), nullable=False)
    package_downloaded = Column(DateTime)
    userid = Column(String(32), nullable=False, index=True)
    participant_type = Column(String(64), nullable=False)
    alma_location = Column(String(32))
    keyword_codes = Column(String(256))
    score_mean = Column(Numeric(8, 6))
    score_std_dev = Column(Numeric(8, 6))
    max_workload = Column(Numeric(3, 0, asdecimal=False))

    cycle = relationship("Ph1mCycle")


class Ph1mProposal(Base):
    __tablename__ = "ph1m_proposal"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    cycle_id = Column(ForeignKey("ALMA.ph1m_cycle.id"), nullable=False)
    archive_uid = Column(String(255), nullable=False, unique=True)
    panel_id_back = Column(Numeric(38, 0, asdecimal=False), index=True)
    cancelled = Column(String(1), server_default=text("'F'"))
    new_integration_time_hours = Column(Numeric(8, 2), server_default=text("0"))
    new_scientific_category = Column(String(16))
    needs_more_time = Column(String(1), server_default=text("'F'"))
    arp_score = Column(Float)
    arp_avg_score = Column(Float)
    arp_std_dev_scores = Column(Float)
    aprc_letter_grade = Column(String(1))
    aprc_rank = Column(Numeric(scale=0, asdecimal=False))
    aprc_original_rank = Column(Numeric(scale=0, asdecimal=False))
    arp_work_progress = Column(String(10))
    aprc_work_progress = Column(String(10), server_default=text("'NOT_SEEN'"))
    feasible = Column(String(8))
    dc_score = Column(Float)
    dc_letter_grade = Column(String(1))
    dc_rank = Column(Numeric(scale=0, asdecimal=False))
    dc_original_rank = Column(Numeric(scale=0, asdecimal=False))
    dc_rank_changed = Column(String(1), server_default=text("'F'"))
    dc_work_progress = Column(String(10), server_default=text("'NOT_SEEN'"))
    arp_rank = Column(Numeric(scale=0, asdecimal=False))
    update_date = Column(DateTime)
    update_userid = Column(String(32))
    triage_flag = Column(String(1), server_default=text("'A'"))
    arp_vote_avg = Column(Float)
    arp_vote_std_dev = Column(Float)
    arp_nr_votes = Column(Numeric(scale=0, asdecimal=False), server_default=text("0"))
    arp_original_rank = Column(Numeric(scale=0, asdecimal=False))
    arp_rank_manually_changed = Column(String(1), server_default=text("'F'"))
    aprc_rank_manually_changed = Column(String(1), server_default=text("'F'"))
    arp_score_normalized = Column(Float, server_default=text("999999"))
    arp_rank_normalized = Column(Float, server_default=text("999999"))
    duplication_flags = Column(String(250))
    arp_comment = Column(String(4000))
    aprc_comment = Column(String(4000))
    new_time_set_by_sg_descope = Column(String(1))
    arp_prop_strength = Column(String(4000))
    arp_prop_weakness = Column(String(4000))
    arp_prop_improvement = Column(String(4000))
    arp_prop_evaluation = Column(String(4000))
    aprc_prop_strength = Column(String(4000))
    aprc_prop_weakness = Column(String(4000))
    aprc_prop_improvement = Column(String(4000))
    aprc_prop_evaluation = Column(String(4000))
    ta_comment_to_pi = Column(String(4000))
    arp_grouping_flag = Column(Numeric(2, 0, asdecimal=False))
    non_standard_twelve_m_hours = Column(Numeric(8, 2))
    manual_flag = Column(String(255))
    new_keywordcode1 = Column(String(4))
    new_keywordcode2 = Column(String(4))

    cycle = relationship("Ph1mCycle")


class Ph1mProposalAuthor(Base):
    __tablename__ = "ph1m_proposal_author"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    proposal_id = Column(ForeignKey("ALMA.ph1m_proposal.id"), nullable=False, index=True)
    sequence = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    userid = Column(String(32), nullable=False, index=True)
    institution_id = Column(Numeric(asdecimal=False))
    author_type = Column(String(4))

    proposal = relationship("Ph1mProposal")


class Ph1mProposalScienceGoal(Base):
    __tablename__ = "ph1m_proposal_science_goal"
    __table_args__ = (Index("ph1m_psg_unique", "proposal_id", "sg_name", unique=True), {"schema": "ALMA"})

    id = Column(Numeric(19, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(19, 0, asdecimal=False), nullable=False)
    proposal_id = Column(ForeignKey("ALMA.ph1m_proposal.id"), nullable=False, index=True)
    sg_name = Column(String(512), nullable=False)
    sg_mode = Column(String(10), nullable=False)
    sg_12m_hours = Column(Numeric(8, 2))
    sg_aca_hours = Column(Numeric(8, 2))
    sg_7m_hours = Column(Numeric(8, 2))
    sg_tp_hours = Column(Numeric(8, 2))

    proposal = relationship("Ph1mProposal")


class Ph1mRecommScienceGoal(Base):
    __tablename__ = "ph1m_recomm_science_goal"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    recommendation_id = Column(ForeignKey("ALMA.ph1m_recommendation.id"), nullable=False)
    science_goal_name = Column(String(512))
    descope = Column(String(1))
    proposal_science_goal_id = Column(ForeignKey("ALMA.ph1m_proposal_science_goal.id"), index=True)

    proposal_science_goal = relationship("Ph1mProposalScienceGoal")
    recommendation = relationship("Ph1mRecommendation")


class Ph1mRecommendation(Base):
    __tablename__ = "ph1m_recommendation"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    recommendation_type = Column(String(32), nullable=False)
    aprc_flag = Column(String(1), server_default=text("null"))
    dc_flag = Column(String(1), server_default=text("'F'"))
    proposal_id = Column(ForeignKey("ALMA.ph1m_proposal.id"))
    update_date = Column(DateTime)
    update_userid = Column(String(32))
    duplicate_proposal_id = Column(ForeignKey("ALMA.ph1m_proposal.id"))
    arc_seconds = Column(Numeric(8, 2))
    text = Column(String(4000))
    text_from_aprc = Column(String(4000))
    prop_recom_uid = Column(Numeric(19, 0, asdecimal=False))

    duplicate_proposal = relationship(
        "Ph1mProposal", primaryjoin="Ph1mRecommendation.duplicate_proposal_id == Ph1mProposal.id"
    )
    proposal = relationship("Ph1mProposal", primaryjoin="Ph1mRecommendation.proposal_id == Ph1mProposal.id")


t_piv_account = Table(
    "piv_account",
    metadata,
    Column("account_id", String(32), nullable=False),
    Column("request_handler_id", Numeric(22, 0, asdecimal=False)),
    Column("firstname", String(256), nullable=False),
    Column("lastname", String(256), nullable=False),
    Column("initials", String(256)),
    Column("preferredarc", String(32)),
    Column("email", String(256)),
    Column("executive", String(5), nullable=False),
    Column("instno", Numeric(38, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


t_piv_execblock = Table(
    "piv_execblock",
    metadata,
    Column("execblockuid", String(32), nullable=False),
    Column("qa0status", String(32), nullable=False),
    Column("aquastarttime", DateTime),
    Column("aquaendtime", DateTime),
    Column("starttime", DateTime),
    Column("endtime", DateTime),
    Column("schedblockuid", String(32)),
    Column("obsprojectpi", String(20)),
    Column("representativefrequency", Numeric(15, 10)),
    Column("obsprojectuid", String(32)),
    schema="ALMA",
)


t_piv_obs_project = Table(
    "piv_obs_project",
    metadata,
    Column("code", String(64)),
    Column("pi", String(64)),
    Column("contact_account_id", String(64)),
    Column("firstname", String(256)),
    Column("lastname", String(256)),
    Column("contact_firstname", String(256)),
    Column("contact_lastname", String(256)),
    Column("title", String(256)),
    Column("prj_version", String(32)),
    Column("prj_letter_grade", String(2)),
    Column("prj_grade", String(1)),
    Column("rank", Numeric(8, 0, asdecimal=False)),
    Column("obs_project_id", String(64), nullable=False),
    Column("domain_entity_state", String(32), nullable=False),
    schema="ALMA",
)


t_piv_obs_proposals = Table(
    "piv_obs_proposals",
    metadata,
    Column("code", String(64)),
    Column("title", String(256)),
    Column("creation_date", String(23)),
    Column("priority_flag", String(2)),
    Column("associatedexec", String(32)),
    Column("scientific_category", String(32)),
    Column("scientific_category_string", String(4000)),
    Column("abstract_text", Text),
    Column("cycle", String(32)),
    Column("proposal_type", String(32)),
    Column("projectuid", String(32), nullable=False),
    Column("archiveuid", String(33), nullable=False),
    Column("consensus_report", String(4000)),
    Column("consensus_report1", String(4000)),
    Column("project_state", String(32), nullable=False),
    schema="ALMA",
)


t_piv_obs_unit_set = Table(
    "piv_obs_unit_set",
    metadata,
    Column("obs_project_id", String(64), nullable=False),
    Column("status_entity_id", String(64), nullable=False),
    Column("parent_obs_unit_set_status_id", String(64)),
    Column("domain_entity_state", String(32), nullable=False),
    Column("partid", String(64)),
    Column("name", String(256)),
    Column("requestedarray", String(8)),
    Column("qa2status", String(8)),
    schema="ALMA",
)


t_piv_proposal_authors = Table(
    "piv_proposal_authors",
    metadata,
    Column("account_id", String(32), nullable=False),
    Column("firstname", String(256), nullable=False),
    Column("lastname", String(256), nullable=False),
    Column("initials", String(256)),
    Column("sequence", Numeric(22, 0, asdecimal=False), nullable=False),
    Column("author_type", String(4), nullable=False),
    Column("project_uid", String(32), nullable=False),
    schema="ALMA",
)


t_piv_sched_block = Table(
    "piv_sched_block",
    metadata,
    Column("obs_project_id", String(64), nullable=False),
    Column("status_entity_id", String(64), nullable=False),
    Column("parent_obs_unit_set_status_id", String(64), nullable=False),
    Column("domain_entity_state", String(32), nullable=False),
    Column("flags", String(200)),
    Column("sched_block_uid", String(64), nullable=False),
    Column("sb_name", String(32)),
    Column("totalestimatedtime", String(32)),
    Column("totalestimatedtimeunits", String(32)),
    Column("ref_ra", String(32)),
    Column("ref_dec", String(32)),
    Column("execution_count", String(32)),
    Column("qa2status", String(8)),
    Column("nominalconfiguration", String(32)),
    schema="ALMA",
)


t_piv_science_goal = Table(
    "piv_science_goal",
    metadata,
    Column("project_uid", String(33), nullable=False),
    Column("project_code", String(64)),
    Column("science_goal_name", String(256)),
    Column("ous_part_id", String(256)),
    Column("ous_status_uid", String(64)),
    Column("sb_uid", String(64)),
    schema="ALMA",
)


t_piv_science_goal_ous_subview = Table(
    "piv_science_goal_ous_subview",
    metadata,
    Column("ous_status_uid", String(64), nullable=False),
    Column("project_uid", String(64), nullable=False),
    Column("ous_part_id", String(64)),
    Column("sb_uid", String(64), nullable=False),
    schema="ALMA",
)


t_piv_science_goal_subview = Table(
    "piv_science_goal_subview",
    metadata,
    Column("project_uid", String(33), nullable=False),
    Column("project_code", String(64)),
    Column("science_goal_name", String(256)),
    Column("ous_part_id", String(256)),
    schema="ALMA",
)


t_piv_state_changes = Table(
    "piv_state_changes",
    metadata,
    Column("state_changes_id", Numeric(32, 0, asdecimal=False), nullable=False),
    Column("status_entity_id", String(64), nullable=False),
    Column("domain_entity_id", String(64), nullable=False),
    Column("domain_entity_state", String(32), nullable=False),
    Column("timestamp", DateTime, nullable=False),
    Column("entity_type", String(3), nullable=False),
    Column("domain_part_id", String(64)),
    Column("child_sb_uid", String(64)),
    schema="ALMA",
)


class PrjAttachment(Base):
    __tablename__ = "prj_attachment"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    obs_project_archive_uid = Column(String(64), nullable=False)
    auth_account_id = Column(String(64), nullable=False)
    created = Column(DateTime, nullable=False)
    filename = Column(String(256), nullable=False)
    contents = Column(LargeBinary, nullable=False)


class PrjComment(Base):
    __tablename__ = "prj_comment"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    obs_project_archive_uid = Column(String(64), nullable=False)
    auth_account_id = Column(String(64), nullable=False)
    created = Column(DateTime, nullable=False)
    updated = Column(DateTime)
    comment_text = Column(String(3500), nullable=False)


class PrjOperation(Base):
    __tablename__ = "prj_operations"
    __table_args__ = {"schema": "ALMA"}

    obs_project_archive_uid = Column(String(64), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    p2g_account_id = Column(String(64))
    contact_account_id = Column(String(64))
    scops_ticket = Column(String(8))
    recipients = Column(String(512))
    helpdesk_ticket = Column(String(8))


class ProjectCode(Base):
    __tablename__ = "project_codes"
    __table_args__ = {"schema": "ALMA"}

    year = Column(String(4), primary_key=True, nullable=False)
    period = Column(String(64), primary_key=True, nullable=False)
    sequence = Column(Numeric(asdecimal=False), nullable=False)
    type_code = Column(String(3), primary_key=True, nullable=False)


t_project_codes_bkp = Table(
    "project_codes_bkp",
    metadata,
    Column("year", String(4), nullable=False),
    Column("period", String(64), nullable=False),
    Column("sequence", Numeric(asdecimal=False), nullable=False),
    Column("type_code", String(3), nullable=False),
    schema="ALMA",
)


t_proposal = Table(
    "proposal",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("cycle_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("archive_uid", String(255), nullable=False),
    Column("panel_id_back", Numeric(38, 0, asdecimal=False)),
    Column("cancelled", String(1)),
    Column("new_integration_time_hours", Numeric(8, 2)),
    Column("new_scientific_category", String(16)),
    Column("needs_more_time", String(1)),
    Column("arp_score", Float),
    Column("arp_avg_score", Float),
    Column("arp_std_dev_scores", Float),
    Column("aprc_letter_grade", String(1)),
    Column("aprc_rank", Numeric(scale=0, asdecimal=False)),
    Column("aprc_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("arp_work_progress", String(10)),
    Column("aprc_work_progress", String(10)),
    Column("feasible", String(8)),
    Column("dc_score", Float),
    Column("dc_letter_grade", String(1)),
    Column("dc_rank", Numeric(scale=0, asdecimal=False)),
    Column("dc_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("dc_rank_changed", String(1)),
    Column("dc_work_progress", String(10)),
    Column("arp_rank", Numeric(scale=0, asdecimal=False)),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("triage_flag", String(1)),
    Column("arp_vote_avg", Float),
    Column("arp_vote_std_dev", Float),
    Column("arp_nr_votes", Numeric(scale=0, asdecimal=False)),
    Column("arp_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("arp_rank_manually_changed", String(1)),
    Column("aprc_rank_manually_changed", String(1)),
    Column("arp_score_normalized", Float),
    Column("arp_rank_normalized", Float),
    Column("duplication_flags", String(250)),
    Column("arp_comment", String(4000)),
    Column("aprc_comment", String(4000)),
    Column("new_time_set_by_sg_descope", String(1)),
    Column("arp_prop_strength", String(4000)),
    Column("arp_prop_weakness", String(4000)),
    Column("arp_prop_improvement", String(4000)),
    Column("arp_prop_evaluation", String(4000)),
    Column("aprc_prop_strength", String(4000)),
    Column("aprc_prop_weakness", String(4000)),
    Column("aprc_prop_improvement", String(4000)),
    Column("aprc_prop_evaluation", String(4000)),
    Column("ta_comment_to_pi", String(4000)),
    Column("arp_grouping_flag", Numeric(2, 0, asdecimal=False)),
    Column("non_standard_twelve_m_hours", Numeric(8, 2)),
    schema="ALMA",
)


t_proposal12_back = Table(
    "proposal12_back",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("cycle_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("archive_uid", String(255), nullable=False),
    Column("alma_review_panel_id", Numeric(38, 0, asdecimal=False)),
    Column("cancelled", String(1)),
    Column("new_integration_time_hours", Numeric(8, 2)),
    Column("new_scientific_category", String(16)),
    Column("needs_more_time", String(1)),
    Column("arp_score", Float),
    Column("arp_comment", Text),
    Column("arp_avg_score", Float),
    Column("arp_std_dev_scores", Float),
    Column("aprc_comment", Text),
    Column("aprc_letter_grade", String(1)),
    Column("aprc_rank", Numeric(scale=0, asdecimal=False)),
    Column("aprc_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("arp_work_progress", String(10)),
    Column("aprc_work_progress", String(10)),
    Column("feasible", String(8)),
    Column("dc_score", Float),
    Column("dc_letter_grade", String(1)),
    Column("dc_rank", Numeric(scale=0, asdecimal=False)),
    Column("dc_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("dc_rank_changed", String(1)),
    Column("dc_work_progress", String(10)),
    Column("arp_rank", Numeric(scale=0, asdecimal=False)),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("triage_flag", String(1)),
    Column("arp_vote_avg", Float),
    Column("arp_vote_std_dev", Float),
    Column("arp_nr_votes", Numeric(scale=0, asdecimal=False)),
    Column("arp_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("arp_rank_manually_changed", String(1)),
    Column("aprc_rank_manually_changed", String(1)),
    Column("arp_score_normalized", Float),
    Column("arp_rank_normalized", Float),
    Column("duplication_flags", String(250)),
    schema="ALMA",
)


t_proposal_author = Table(
    "proposal_author",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("proposal_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("sequence", Numeric(scale=0, asdecimal=False), nullable=False),
    Column("userid", String(32), nullable=False),
    Column("institution_id", Numeric(asdecimal=False)),
    schema="ALMA",
)


t_proposal_cycle2_backup = Table(
    "proposal_cycle2_backup",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("cycle_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("archive_uid", String(255), nullable=False),
    Column("alma_review_panel_id", Numeric(38, 0, asdecimal=False)),
    Column("cancelled", String(1)),
    Column("new_integration_time_hours", Numeric(8, 2)),
    Column("new_scientific_category", String(16)),
    Column("needs_more_time", String(1)),
    Column("arp_score", Float),
    Column("arp_comment", Text),
    Column("arp_avg_score", Float),
    Column("arp_std_dev_scores", Float),
    Column("aprc_comment", Text),
    Column("aprc_letter_grade", String(1)),
    Column("aprc_rank", Numeric(scale=0, asdecimal=False)),
    Column("aprc_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("arp_work_progress", String(10)),
    Column("aprc_work_progress", String(10)),
    Column("feasible", String(8)),
    Column("dc_score", Float),
    Column("dc_letter_grade", String(1)),
    Column("dc_rank", Numeric(scale=0, asdecimal=False)),
    Column("dc_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("dc_rank_changed", String(1)),
    Column("dc_work_progress", String(10)),
    Column("arp_rank", Numeric(scale=0, asdecimal=False)),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("triage_flag", String(1)),
    Column("arp_vote_avg", Float),
    Column("arp_vote_std_dev", Float),
    Column("arp_nr_votes", Numeric(scale=0, asdecimal=False)),
    Column("arp_original_rank", Numeric(scale=0, asdecimal=False)),
    Column("arp_rank_manually_changed", String(1)),
    Column("aprc_rank_manually_changed", String(1)),
    Column("arp_score_normalized", Float),
    Column("arp_rank_normalized", Float),
    Column("duplication_flags", String(250)),
    schema="ALMA",
)


t_recommendation = Table(
    "recommendation",
    metadata,
    Column("id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("version", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("recommendation_type", String(32), nullable=False),
    Column("aprc_flag", String(1)),
    Column("dc_flag", String(1)),
    Column("proposal_id", Numeric(38, 0, asdecimal=False)),
    Column("update_date", DateTime),
    Column("update_userid", String(32)),
    Column("duplicate_proposal_id", Numeric(38, 0, asdecimal=False)),
    Column("arc_seconds", Numeric(8, 2)),
    Column("text", String(4000)),
    Column("text_from_aprc", String(4000)),
    schema="ALMA",
)


class Role(Base):
    __tablename__ = "role"
    __table_args__ = (Index("role_name_unique", "application", "name", unique=True), {"schema": "ALMA"})

    role_no = Column(Numeric(38, 0, asdecimal=False), primary_key=True)
    version = Column(Numeric(38, 0, asdecimal=False), nullable=False)
    application = Column(String(32), nullable=False)
    name = Column(String(32), nullable=False)
    parent_role = Column(ForeignKey("ALMA.role.role_no"))

    parent = relationship("Role", remote_side=[role_no])


class SchedBlockStatu(Base):
    __tablename__ = "sched_block_status"
    __table_args__ = (Index("sched_block_status_idx", "obs_project_id", "domain_entity_id"), {"schema": "ALMA"})

    status_entity_id = Column(String(64), primary_key=True)
    domain_entity_id = Column(ForeignKey("ALMA.xml_schedblock_entities.archive_uid"), nullable=False, index=True)
    domain_entity_state = Column(String(32), nullable=False)
    parent_obs_unit_set_status_id = Column(String(64), nullable=False, index=True)
    obs_project_status_id = Column(String(64), nullable=False)
    obs_project_id = Column(ForeignKey("ALMA.xml_obsproject_entities.archive_uid"), nullable=False)
    total_required_time_in_sec = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    total_used_time_in_sec = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    xml = Column(NullType)
    flags = Column(String(200), index=True)

    domain_entity = relationship("XmlSchedblockEntity")
    obs_project = relationship("XmlObsprojectEntity")


class SchemaVersion(Base):
    __tablename__ = "schema_version"
    __table_args__ = {"schema": "ALMA"}

    name = Column(String(32), primary_key=True)
    value = Column(String(32), nullable=False)


class ShiftlogEntry(Base):
    __tablename__ = "shiftlog_entries"
    __table_args__ = {"schema": "ALMA"}

    se_id = Column(Numeric(19, 0, asdecimal=False), primary_key=True)
    se_type = Column(Numeric(10, 0, asdecimal=False), nullable=False, index=True)
    se_subject = Column(String(256))
    se_timestamp = Column(DateTime, nullable=False, unique=True)
    se_author = Column(String(32))
    se_start = Column(DateTime, index=True)
    se_project_code = Column(String(128), index=True)
    se_sb_code = Column(String(128))
    se_sb_id = Column(String(32), index=True)
    se_eb_uid = Column(String(32))
    se_location = Column(String(32), index=True)
    se_status = Column(String(32))
    se_calibration = Column(String(1))
    se_qa0flag = Column(String(32))
    se_archiving_status = Column(String(32))
    se_test_activity = Column(String(64))
    se_ispowercut = Column(String(1))
    se_pcrecoveryend = Column(DateTime)
    se_pcrecoverystart = Column(DateTime)
    se_wrecoveryend12m = Column(DateTime)
    se_wrecoveryend7m = Column(DateTime)
    se_wrecoveryendtp = Column(DateTime)
    se_wrecoverystart12m = Column(DateTime)
    se_wrecoverystart7m = Column(DateTime)
    se_wrecoverystarttp = Column(DateTime)
    se_arrayentry_id = Column(ForeignKey("ALMA.shiftlog_entries.se_id"))
    se_arrayname = Column(String(10))
    se_arraytype = Column(String(9))
    se_arrayfamily = Column(String(11))
    se_correlatortype = Column(String(4))
    se_photonicreferencename = Column(String(18))
    se_almabuild = Column(String(70))
    se_downtimetype = Column(String(9))
    se_mainactivity = Column(String(100))
    se_mainantennaname = Column(String(50))
    se_mainresponsible = Column(String(50))
    se_bandname = Column(String(10))
    se_executive = Column(String(23))
    se_obsprojectname = Column(String(256))
    se_obsprojectpi = Column(String(20))
    se_obsprojectversion = Column(String(20))
    se_acsversion = Column(String(12))
    se_shiftactivity = Column(String(11))
    se_dashboardantennasavailable = Column(String(2))
    se_dashboardantennasdelivered = Column(String(2))
    se_dashboardantennaspresent = Column(String(2))
    se_pwv = Column(Numeric(18, 10))
    se_reprfrequency = Column(Numeric(15, 10))
    se_test = Column(String(1))

    se_arrayentry = relationship("ShiftlogEntry", remote_side=[se_id])


t_shiftlog_entries_bak = Table(
    "shiftlog_entries_bak",
    metadata,
    Column("se_id", Numeric(19, 0, asdecimal=False)),
    Column("se_type", Numeric(10, 0, asdecimal=False), nullable=False),
    Column("se_subject", String(256)),
    Column("se_timestamp", DateTime, nullable=False),
    Column("se_author", String(32)),
    Column("se_start", DateTime),
    Column("se_project_code", String(128)),
    Column("se_sb_code", String(128)),
    Column("se_sb_id", String(32)),
    Column("se_eb_uid", String(32)),
    Column("se_location", String(32)),
    Column("se_status", String(32)),
    Column("se_calibration", String(1)),
    Column("se_qa0flag", String(10)),
    Column("se_archiving_status", String(32)),
    Column("se_test_activity", String(64)),
    Column("se_ispowercut", String(1)),
    Column("se_pcrecoveryend", DateTime),
    Column("se_pcrecoverystart", DateTime),
    Column("se_wrecoveryend12m", DateTime),
    Column("se_wrecoveryend7m", DateTime),
    Column("se_wrecoveryendtp", DateTime),
    Column("se_wrecoverystart12m", DateTime),
    Column("se_wrecoverystart7m", DateTime),
    Column("se_wrecoverystarttp", DateTime),
    Column("se_arrayentry_id", Numeric(10, 0, asdecimal=False)),
    Column("se_arrayname", String(10)),
    Column("se_arraytype", String(9)),
    Column("se_arrayfamily", String(11)),
    Column("se_correlatortype", String(4)),
    Column("se_photonicreferencename", String(18)),
    Column("se_almabuild", String(70)),
    Column("se_downtimetype", String(9)),
    Column("se_mainactivity", String(100)),
    Column("se_mainantennaname", String(50)),
    Column("se_mainresponsible", String(50)),
    Column("se_bandname", String(10)),
    Column("se_executive", String(20)),
    Column("se_obsprojectname", String(256)),
    Column("se_obsprojectpi", String(20)),
    Column("se_obsprojectversion", String(20)),
    Column("se_acsversion", String(12)),
    Column("se_shiftactivity", String(11)),
    Column("se_dashboardantennasavailable", String(2)),
    Column("se_dashboardantennasdelivered", String(2)),
    Column("se_dashboardantennaspresent", String(2)),
    Column("se_pwv", Numeric(18, 10)),
    Column("se_reprfrequency", Numeric(15, 10)),
    schema="ALMA",
)


class ShiftlogReply(Base):
    __tablename__ = "shiftlog_reply"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(19, 0, asdecimal=False), primary_key=True)
    reply_comment = Column(Text)
    timestamp = Column(DateTime)
    author = Column(String(32))
    entry_id = Column(ForeignKey("ALMA.shiftlog_entries.se_id"), nullable=False, index=True)

    entry = relationship("ShiftlogEntry")


t_shiftlog_reply_bak = Table(
    "shiftlog_reply_bak",
    metadata,
    Column("id", Numeric(19, 0, asdecimal=False)),
    Column("reply_comment", Text),
    Column("timestamp", DateTime),
    Column("author", String(32)),
    Column("entry_id", Numeric(19, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


class ShiftlogSbUsedTime(Base):
    __tablename__ = "shiftlog_sb_used_time"
    __table_args__ = {"schema": "ALMA"}

    sb_uid = Column(String(33), primary_key=True)
    used_time = Column(Numeric(asdecimal=False), nullable=False)


t_shiftlog_sb_used_time_bak = Table(
    "shiftlog_sb_used_time_bak",
    metadata,
    Column("sb_uid", String(33), nullable=False),
    Column("used_time", Numeric(asdecimal=False), nullable=False),
    schema="ALMA",
)


class SlogAllocatedTimeInterval(Base):
    __tablename__ = "slog_allocated_time_interval"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(10, 0, asdecimal=False), primary_key=True)
    entry_id = Column(ForeignKey("ALMA.shiftlog_entries.se_id"), nullable=False)
    arrayfamily = Column(String(11), nullable=False)
    starttime = Column(DateTime, nullable=False)
    endtime = Column(DateTime)

    entry = relationship("ShiftlogEntry")


class SlogEntryAttach(Base):
    __tablename__ = "slog_entry_attach"
    __table_args__ = {"schema": "ALMA"}

    slog_attach_id = Column(Numeric(19, 0, asdecimal=False), primary_key=True)
    slog_attach_name = Column(String(256))
    slog_attach_contents = Column(LargeBinary)
    slog_se_id = Column(ForeignKey("ALMA.shiftlog_entries.se_id"), nullable=False, index=True)

    slog_se = relationship("ShiftlogEntry")


t_slog_entry_attach_bak = Table(
    "slog_entry_attach_bak",
    metadata,
    Column("slog_attach_id", Numeric(19, 0, asdecimal=False)),
    Column("slog_attach_name", String(256)),
    Column("slog_attach_contents", LargeBinary),
    Column("slog_se_id", Numeric(19, 0, asdecimal=False), nullable=False),
    schema="ALMA",
)


class SlogEntryAttr(Base):
    __tablename__ = "slog_entry_attr"
    __table_args__ = {"schema": "ALMA"}

    slog_attr_id = Column(Numeric(19, 0, asdecimal=False), primary_key=True)
    slog_attr_type = Column(Numeric(10, 0, asdecimal=False), nullable=False)
    slog_attr_value = Column(String(256))
    slog_se_id = Column(ForeignKey("ALMA.shiftlog_entries.se_id"), index=True)

    slog_se = relationship("ShiftlogEntry")


t_slog_entry_attr_bak = Table(
    "slog_entry_attr_bak",
    metadata,
    Column("slog_attr_id", Numeric(19, 0, asdecimal=False), nullable=False),
    Column("slog_attr_type", Numeric(10, 0, asdecimal=False), nullable=False),
    Column("slog_attr_value", String(256)),
    Column("slog_se_id", Numeric(19, 0, asdecimal=False)),
    schema="ALMA",
)


class StateChange(Base):
    __tablename__ = "state_changes"
    __table_args__ = (Index("id_perf_schs", "status_entity_id", "domain_entity_state"), {"schema": "ALMA"})

    state_changes_id = Column(Numeric(32, 0, asdecimal=False), primary_key=True)
    status_entity_id = Column(String(64), nullable=False, index=True)
    domain_entity_id = Column(String(64), nullable=False, index=True)
    domain_entity_state = Column(String(32), nullable=False)
    timestamp = Column(DateTime, nullable=False, index=True)
    location = Column(String(3), nullable=False)
    user_id = Column(String(64), nullable=False)
    subsystem = Column(String(32), nullable=False)
    info = Column(String(2000))
    entity_type = Column(String(3), nullable=False)
    domain_part_id = Column(String(64))
    flags = Column(String(200), index=True)


t_submission_service_account = Table(
    "submission_service_account",
    metadata,
    Column("account_id", String(32), nullable=False),
    Column("password_digest", String(256)),
    schema="ALMA",
)


t_submission_service_roles = Table(
    "submission_service_roles", metadata, Column("account_id", String(32)), Column("name", String(66)), schema="ALMA"
)


t_submission_service_roles_orig = Table(
    "submission_service_roles_orig",
    metadata,
    Column("account_id", String(32), nullable=False),
    Column("name", String(66)),
    schema="ALMA",
)


t_test_23 = Table("test_23", metadata, Column("a", Numeric(scale=0, asdecimal=False)), schema="ALMA")


t_test_24 = Table("test_24", metadata, Column("a", Numeric(scale=0, asdecimal=False)), schema="ALMA")


class TrustedAccount(Base):
    __tablename__ = "trusted_accounts"
    __table_args__ = (
        Index("trustedacc_unique", "alma_account_id", "trusted_account_id", "trusted_institution", unique=True),
        {"schema": "ALMA"},
    )

    id = Column(Numeric(asdecimal=False), primary_key=True)
    alma_account_id = Column(ForeignKey("ALMA.account.account_id"), nullable=False)
    trusted_account_id = Column(String(32), nullable=False)
    trusted_institution = Column(String(10), nullable=False)

    alma_account = relationship("Account")


t_uid_lookup = Table(
    "uid_lookup", metadata, Column("archive_uid", String(33)), Column("schemaname", String(30)), schema="ALMA"
)


class UserDemographic(Base):
    __tablename__ = "user_demographics"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    version = Column(Numeric(asdecimal=False))
    account_id = Column(ForeignKey("ALMA.account.account_id"), nullable=False)
    observing_mm_submm = Column(String(32), nullable=False)
    observing_ir = Column(String(32), nullable=False)
    observing_radio = Column(String(32), nullable=False)
    observing_theory_modeling = Column(String(32), nullable=False)
    observing_xray = Column(String(32), nullable=False)
    professional_status = Column(String(32), nullable=False)
    techniques_interferometry = Column(String(32), nullable=False)
    techniques_total_power = Column(String(32), nullable=False)
    year_of_phd = Column(Numeric(scale=0, asdecimal=False), nullable=False)
    last_update = Column(DateTime, nullable=False)
    observing_uv_optical = Column(String(32), nullable=False)

    account = relationship("Account")


t_v_region_props = Table(
    "v_region_props",
    metadata,
    Column("v_proposal_id", Numeric(38, 0, asdecimal=False), nullable=False),
    Column("v_pi_executive", String(32)),
    Column("v_aprc_letter_grade", String(1)),
    Column("v_aprc_rank", Numeric(scale=0, asdecimal=False)),
    Column("v_arp_score", Float),
    Column("regional_rank", Numeric(asdecimal=False)),
    Column("nr_props_region", Numeric(asdecimal=False)),
    schema="ALMA",
)


t_v_schedblock_polarisation = Table(
    "v_schedblock_polarisation",
    metadata,
    Column("archive_uid", String(33), nullable=False),
    Column("products", String(32), nullable=False),
    schema="ALMA",
)


t_watchdog = Table("watchdog", metadata, Column("field", DateTime), Column("site", String(5)), schema="ALMA")


class XmlAcapolarizationEntity(Base):
    __tablename__ = "xml_acapolarization_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAccummodeEntity(Base):
    __tablename__ = "xml_accummode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAcsalarmmessageEntity(Base):
    __tablename__ = "xml_acsalarmmessage_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAcscommandcenterpEntity(Base):
    __tablename__ = "xml_acscommandcenterp_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAcscommandcentertEntity(Base):
    __tablename__ = "xml_acscommandcentert_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAcserrorEntity(Base):
    __tablename__ = "xml_acserror_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAcslogtsEntity(Base):
    __tablename__ = "xml_acslogts_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAddressEntity(Base):
    __tablename__ = "xml_address_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAlarmsystemstatsEntity(Base):
    __tablename__ = "xml_alarmsystemstats_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAlmaradiometertabEntity(Base):
    __tablename__ = "xml_almaradiometertab_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAnnotationtableEntity(Base):
    __tablename__ = "xml_annotationtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAntennamakeEntity(Base):
    __tablename__ = "xml_antennamake_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAntennamotionpattEntity(Base):
    __tablename__ = "xml_antennamotionpatt_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAntennatableEntity(Base):
    __tablename__ = "xml_antennatable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAntennatypeEntity(Base):
    __tablename__ = "xml_antennatype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAsdmEntity(Base):
    __tablename__ = "xml_asdm_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAsdmbinarytableEntity(Base):
    __tablename__ = "xml_asdmbinarytable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAsiconfigurationEntity(Base):
    __tablename__ = "xml_asiconfiguration_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAsimessageEntity(Base):
    __tablename__ = "xml_asimessage_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAssociatedcalnatuEntity(Base):
    __tablename__ = "xml_associatedcalnatu_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAssociatedfieldnaEntity(Base):
    __tablename__ = "xml_associatedfieldna_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAtmphasecorrectioEntity(Base):
    __tablename__ = "xml_atmphasecorrectio_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlAxisnameEntity(Base):
    __tablename__ = "xml_axisname_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlBasebandnameEntity(Base):
    __tablename__ = "xml_basebandname_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlBaselinereferenceEntity(Base):
    __tablename__ = "xml_baselinereference_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlBeamtableEntity(Base):
    __tablename__ = "xml_beamtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlBinarydataflagsEntity(Base):
    __tablename__ = "xml_binarydataflags_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlBulktestEntity(Base):
    __tablename__ = "xml_bulktest_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalamplitableEntity(Base):
    __tablename__ = "xml_calamplitable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalappphasetableEntity(Base):
    __tablename__ = "xml_calappphasetable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalatmospheretablEntity(Base):
    __tablename__ = "xml_calatmospheretabl_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalbandpasstableEntity(Base):
    __tablename__ = "xml_calbandpasstable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalcurvetableEntity(Base):
    __tablename__ = "xml_calcurvetable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalcurvetypeEntity(Base):
    __tablename__ = "xml_calcurvetype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCaldataoriginEntity(Base):
    __tablename__ = "xml_caldataorigin_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCaldatatableEntity(Base):
    __tablename__ = "xml_caldatatable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCaldelaytableEntity(Base):
    __tablename__ = "xml_caldelaytable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCaldevicetableEntity(Base):
    __tablename__ = "xml_caldevicetable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalfluxtableEntity(Base):
    __tablename__ = "xml_calfluxtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalfocusmodeltablEntity(Base):
    __tablename__ = "xml_calfocusmodeltabl_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalfocustableEntity(Base):
    __tablename__ = "xml_calfocustable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalgaintableEntity(Base):
    __tablename__ = "xml_calgaintable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalholographytablEntity(Base):
    __tablename__ = "xml_calholographytabl_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalibrationconfigEntity(Base):
    __tablename__ = "xml_calibrationconfig_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalibrationdeviceEntity(Base):
    __tablename__ = "xml_calibrationdevice_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalibrationfunctiEntity(Base):
    __tablename__ = "xml_calibrationfuncti_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalibrationmodeEntity(Base):
    __tablename__ = "xml_calibrationmode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalibrationsetEntity(Base):
    __tablename__ = "xml_calibrationset_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalphasetableEntity(Base):
    __tablename__ = "xml_calphasetable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalpointingmodeltEntity(Base):
    __tablename__ = "xml_calpointingmodelt_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalpointingtableEntity(Base):
    __tablename__ = "xml_calpointingtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalpositiontableEntity(Base):
    __tablename__ = "xml_calpositiontable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalprimarybeamtabEntity(Base):
    __tablename__ = "xml_calprimarybeamtab_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalqueryparameterEntity(Base):
    __tablename__ = "xml_calqueryparameter_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalreductiontableEntity(Base):
    __tablename__ = "xml_calreductiontable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalseeingtableEntity(Base):
    __tablename__ = "xml_calseeingtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCaltypeEntity(Base):
    __tablename__ = "xml_caltype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCalwvrtableEntity(Base):
    __tablename__ = "xml_calwvrtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCommonentityEntity(Base):
    __tablename__ = "xml_commonentity_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCommontypesEntity(Base):
    __tablename__ = "xml_commontypes_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlConfigdescriptionEntity(Base):
    __tablename__ = "xml_configdescription_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCorrelationbitEntity(Base):
    __tablename__ = "xml_correlationbit_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCorrelationmodeEntity(Base):
    __tablename__ = "xml_correlationmode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCorrelatorcalibraEntity(Base):
    __tablename__ = "xml_correlatorcalibra_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCorrelatormodetabEntity(Base):
    __tablename__ = "xml_correlatormodetab_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCorrelatornameEntity(Base):
    __tablename__ = "xml_correlatorname_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlCorrelatortypeEntity(Base):
    __tablename__ = "xml_correlatortype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDatacontentEntity(Base):
    __tablename__ = "xml_datacontent_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDatadescriptiontaEntity(Base):
    __tablename__ = "xml_datadescriptionta_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDatascaleEntity(Base):
    __tablename__ = "xml_datascale_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDelaymodelfixedpaEntity(Base):
    __tablename__ = "xml_delaymodelfixedpa_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDelaymodeltableEntity(Base):
    __tablename__ = "xml_delaymodeltable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDelaymodelvariablEntity(Base):
    __tablename__ = "xml_delaymodelvariabl_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDetectorbandtypeEntity(Base):
    __tablename__ = "xml_detectorbandtype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDifferencetypeEntity(Base):
    __tablename__ = "xml_differencetype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDirectionreferencEntity(Base):
    __tablename__ = "xml_directionreferenc_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDopplerreferencecEntity(Base):
    __tablename__ = "xml_dopplerreferencec_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDopplertableEntity(Base):
    __tablename__ = "xml_dopplertable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlDopplertrackingmoEntity(Base):
    __tablename__ = "xml_dopplertrackingmo_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlEphemeristableEntity(Base):
    __tablename__ = "xml_ephemeristable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlExecblocktableEntity(Base):
    __tablename__ = "xml_execblocktable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlExecconfigEntity(Base):
    __tablename__ = "xml_execconfig_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFeedtableEntity(Base):
    __tablename__ = "xml_feedtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFieldcodeEntity(Base):
    __tablename__ = "xml_fieldcode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFieldtableEntity(Base):
    __tablename__ = "xml_fieldtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFiltermodeEntity(Base):
    __tablename__ = "xml_filtermode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFlagcmdtableEntity(Base):
    __tablename__ = "xml_flagcmdtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFlagtableEntity(Base):
    __tablename__ = "xml_flagtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFluxcalibrationmeEntity(Base):
    __tablename__ = "xml_fluxcalibrationme_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFocusmethodEntity(Base):
    __tablename__ = "xml_focusmethod_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFocusmodeltableEntity(Base):
    __tablename__ = "xml_focusmodeltable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFocustableEntity(Base):
    __tablename__ = "xml_focustable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFreqoffsettableEntity(Base):
    __tablename__ = "xml_freqoffsettable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlFrequencyreferencEntity(Base):
    __tablename__ = "xml_frequencyreferenc_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlGaintrackingtableEntity(Base):
    __tablename__ = "xml_gaintrackingtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlHistorytableEntity(Base):
    __tablename__ = "xml_historytable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlHolographychannelEntity(Base):
    __tablename__ = "xml_holographychannel_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlHolographytableEntity(Base):
    __tablename__ = "xml_holographytable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlIdentifierrangeEntity(Base):
    __tablename__ = "xml_identifierrange_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlInvalidatingcondiEntity(Base):
    __tablename__ = "xml_invalidatingcondi_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlLoggingmiEntity(Base):
    __tablename__ = "xml_loggingmi_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlMaintableEntity(Base):
    __tablename__ = "xml_maintable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlMetaHistory(Base):
    __tablename__ = "xml_meta_history"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True, nullable=False)
    timestamp = Column(DateTime, primary_key=True, nullable=False)
    metacolumn = Column(Numeric(16, 0, asdecimal=False), primary_key=True, nullable=False)
    newvalue = Column(String(64))


t_xml_metainfo = Table(
    "xml_metainfo", metadata, Column("name", String(32)), Column("value", String(128)), schema="ALMA"
)


t_xml_metainfo_backup = Table(
    "xml_metainfo_backup", metadata, Column("name", String(32)), Column("value", String(128)), schema="ALMA"
)


class XmlNamespace(Base):
    __tablename__ = "xml_namespaces"
    __table_args__ = {"schema": "ALMA"}

    prefix = Column(String(16), primary_key=True)
    namespace = Column(String(128))


class XmlNetsidebandEntity(Base):
    __tablename__ = "xml_netsideband_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObsattachmentEntity(Base):
    __tablename__ = "xml_obsattachment_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObservationtableEntity(Base):
    __tablename__ = "xml_observationtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObservingcontrolsEntity(Base):
    __tablename__ = "xml_observingcontrols_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObservingmodeEntity(Base):
    __tablename__ = "xml_observingmode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObsprojectEntity(Base):
    __tablename__ = "xml_obsproject_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObsproposalEntity(Base):
    __tablename__ = "xml_obsproposal_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObsreviewEntity(Base):
    __tablename__ = "xml_obsreview_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlObstooluserprefsEntity(Base):
    __tablename__ = "xml_obstooluserprefs_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlOtconfigurationEntity(Base):
    __tablename__ = "xml_otconfiguration_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlOusstatusEntity(Base):
    __tablename__ = "xml_ousstatus_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPointingmethodEntity(Base):
    __tablename__ = "xml_pointingmethod_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPointingmodelmodeEntity(Base):
    __tablename__ = "xml_pointingmodelmode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPointingmodeltablEntity(Base):
    __tablename__ = "xml_pointingmodeltabl_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPointingtableEntity(Base):
    __tablename__ = "xml_pointingtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPolarizationtableEntity(Base):
    __tablename__ = "xml_polarizationtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPolarizationtypeEntity(Base):
    __tablename__ = "xml_polarizationtype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPositionmethodEntity(Base):
    __tablename__ = "xml_positionmethod_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPositionreferenceEntity(Base):
    __tablename__ = "xml_positionreference_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPreferencesEntity(Base):
    __tablename__ = "xml_preferences_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPrimarybeamdescriEntity(Base):
    __tablename__ = "xml_primarybeamdescri_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPrimitivedatatypeEntity(Base):
    __tablename__ = "xml_primitivedatatype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlProcessorsubtypeEntity(Base):
    __tablename__ = "xml_processorsubtype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlProcessortableEntity(Base):
    __tablename__ = "xml_processortable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlProcessortypeEntity(Base):
    __tablename__ = "xml_processortype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlProjectstatusEntity(Base):
    __tablename__ = "xml_projectstatus_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPsetEntity(Base):
    __tablename__ = "xml_pset_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlPsetdefEntity(Base):
    __tablename__ = "xml_psetdef_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQlalarmsEntity(Base):
    __tablename__ = "xml_qlalarms_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQlatmospheresummaEntity(Base):
    __tablename__ = "xml_qlatmospheresumma_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQlfocussummaryEntity(Base):
    __tablename__ = "xml_qlfocussummary_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQlphasesummaryEntity(Base):
    __tablename__ = "xml_qlphasesummary_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQlpointingsummaryEntity(Base):
    __tablename__ = "xml_qlpointingsummary_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQuicklookdisplayEntity(Base):
    __tablename__ = "xml_quicklookdisplay_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQuicklookdisplayxEntity(Base):
    __tablename__ = "xml_quicklookdisplayx_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQuicklookresultEntity(Base):
    __tablename__ = "xml_quicklookresult_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlQuicklooksummaryEntity(Base):
    __tablename__ = "xml_quicklooksummary_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlRadialvelocityrefEntity(Base):
    __tablename__ = "xml_radialvelocityref_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlReceiverbandEntity(Base):
    __tablename__ = "xml_receiverband_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlReceiversidebandEntity(Base):
    __tablename__ = "xml_receiversideband_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlReceivertableEntity(Base):
    __tablename__ = "xml_receivertable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlRole(Base):
    __tablename__ = "xml_roles"
    __table_args__ = {"schema": "ALMA"}

    rolename = Column(String(64), primary_key=True)


class XmlSbstatusEntity(Base):
    __tablename__ = "xml_sbstatus_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSbsummarytableEntity(Base):
    __tablename__ = "xml_sbsummarytable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSbtypeEntity(Base):
    __tablename__ = "xml_sbtype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlScaletableEntity(Base):
    __tablename__ = "xml_scaletable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlScanintentEntity(Base):
    __tablename__ = "xml_scanintent_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlScantableEntity(Base):
    __tablename__ = "xml_scantable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSchedblockEntity(Base):
    __tablename__ = "xml_schedblock_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False), index=True)
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSchedulermodeEntity(Base):
    __tablename__ = "xml_schedulermode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSchedulingpolicyEntity(Base):
    __tablename__ = "xml_schedulingpolicy_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSchemaEntity(Base):
    __tablename__ = "xml_schema_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    schemaname = Column(String(32), nullable=False)
    version = Column(Numeric(16, 0, asdecimal=False), nullable=False)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(Text)
    schemauid = Column(String(33))
    owner = Column(String(128))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(128))
    writepermissions = Column(String(128))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSchemanamespace(Base):
    __tablename__ = "xml_schemanamespaces"
    __table_args__ = {"schema": "ALMA"}

    schemauid = Column(String(33), primary_key=True, nullable=False)
    prefix = Column(String(16), primary_key=True, nullable=False)


class XmlScipiperequestEntity(Base):
    __tablename__ = "xml_scipiperequest_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlScipiperesultsEntity(Base):
    __tablename__ = "xml_scipiperesults_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSdmdataheaderEntity(Base):
    __tablename__ = "xml_sdmdataheader_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSeeingtableEntity(Base):
    __tablename__ = "xml_seeingtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSidebandprocessinEntity(Base):
    __tablename__ = "xml_sidebandprocessin_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSourcemodelEntity(Base):
    __tablename__ = "xml_sourcemodel_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSourcetableEntity(Base):
    __tablename__ = "xml_sourcetable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSpecialsbEntity(Base):
    __tablename__ = "xml_specialsb_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSpectralresolutioEntity(Base):
    __tablename__ = "xml_spectralresolutio_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSpectralwindowtabEntity(Base):
    __tablename__ = "xml_spectralwindowtab_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False, index=True)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSquarelawdetectorEntity(Base):
    __tablename__ = "xml_squarelawdetector_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlStatetableEntity(Base):
    __tablename__ = "xml_statetable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlStationtableEntity(Base):
    __tablename__ = "xml_stationtable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlStationtypeEntity(Base):
    __tablename__ = "xml_stationtype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlStokesparameterEntity(Base):
    __tablename__ = "xml_stokesparameter_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlStylesheet(Base):
    __tablename__ = "xml_stylesheets"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemaname = Column(String(32), nullable=False)
    schemauid = Column(String(33), nullable=False)
    version_from = Column(Numeric(16, 0, asdecimal=False))
    version_to = Column(Numeric(16, 0, asdecimal=False))
    alma_release = Column(String(32), nullable=False)


class XmlSubscanfieldsourcEntity(Base):
    __tablename__ = "xml_subscanfieldsourc_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSubscanintentEntity(Base):
    __tablename__ = "xml_subscanintent_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSubscanspectralspEntity(Base):
    __tablename__ = "xml_subscanspectralsp_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSubscantableEntity(Base):
    __tablename__ = "xml_subscantable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSwitchcycletableEntity(Base):
    __tablename__ = "xml_switchcycletable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSwitchingmodeEntity(Base):
    __tablename__ = "xml_switchingmode_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSynthprofEntity(Base):
    __tablename__ = "xml_synthprof_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSyscalmethodEntity(Base):
    __tablename__ = "xml_syscalmethod_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSyscaltableEntity(Base):
    __tablename__ = "xml_syscaltable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlSyspowertableEntity(Base):
    __tablename__ = "xml_syspowertable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTestobsprojectEntity(Base):
    __tablename__ = "xml_testobsproject_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTestobsproposalEntity(Base):
    __tablename__ = "xml_testobsproposal_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTestprojectEntity(Base):
    __tablename__ = "xml_testproject_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTestschedblockEntity(Base):
    __tablename__ = "xml_testschedblock_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTestvaluetypesEntity(Base):
    __tablename__ = "xml_testvaluetypes_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTimesamplingEntity(Base):
    __tablename__ = "xml_timesampling_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTimescaleEntity(Base):
    __tablename__ = "xml_timescale_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlTotalpowertableEntity(Base):
    __tablename__ = "xml_totalpowertable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlUpdate(Base):
    __tablename__ = "xml_updates"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(asdecimal=False), primary_key=True)
    asdm_uid = Column(ForeignKey("ALMA.xml_asdm_entities.archive_uid"), nullable=False)
    timestamp = Column(DateTime, nullable=False)
    modified_by = Column(String(32), nullable=False)
    hostname = Column(String(32), nullable=False)
    elaboration = Column(String(4000), nullable=False)
    delta_version = Column(Numeric(5, 0, asdecimal=False), nullable=False)

    xml_asdm_entity = relationship("XmlAsdmEntity")


class XmlUserEntity(Base):
    __tablename__ = "xml_user_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlUserrole(Base):
    __tablename__ = "xml_userroles"
    __table_args__ = {"schema": "ALMA"}

    username = Column(String(64), primary_key=True, nullable=False)
    rolename = Column(String(64), primary_key=True, nullable=False)


class XmlUser(Base):
    __tablename__ = "xml_users"
    __table_args__ = {"schema": "ALMA"}

    username = Column(String(64), primary_key=True)


class XmlValuetypesEntity(Base):
    __tablename__ = "xml_valuetypes_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlWeathertableEntity(Base):
    __tablename__ = "xml_weathertable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlWeighttypeEntity(Base):
    __tablename__ = "xml_weighttype_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlWindowfunctionEntity(Base):
    __tablename__ = "xml_windowfunction_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlWvmcaltableEntity(Base):
    __tablename__ = "xml_wvmcaltable_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class XmlWvrmethodEntity(Base):
    __tablename__ = "xml_wvrmethod_entities"
    __table_args__ = {"schema": "ALMA"}

    archive_uid = Column(String(33), primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    xml = Column(NullType)
    schemauid = Column(String(33), nullable=False)
    owner = Column(String(32))
    deleted = Column(Numeric(1, 0, asdecimal=False))
    readpermissions = Column(String(8))
    writepermissions = Column(String(8))
    hidden = Column(Numeric(1, 0, asdecimal=False))
    dirty = Column(Numeric(1, 0, asdecimal=False))
    virtual = Column(Numeric(1, 0, asdecimal=False))


class Xslt(Base):
    __tablename__ = "xslt"
    __table_args__ = {"schema": "ALMA"}

    id = Column(Numeric(3, 0, asdecimal=False), primary_key=True)
    t_xsl = Column(NullType)


class AquaOusComment(AquaComment):
    __tablename__ = "aqua_ous_comment"
    __table_args__ = {"schema": "ALMA"}

    commentid = Column(ForeignKey("ALMA.aqua_comment.commentid"), primary_key=True)
    obsunitsetid = Column(ForeignKey("ALMA.aqua_ous.obsunitsetid"))
    comment_section = Column(String(8))

    aqua_ou = relationship("AquaOu")


class AquaQa2ReasonHistory(AquaStatusHistory):
    __tablename__ = "aqua_qa2_reason_history"
    __table_args__ = {"schema": "ALMA"}

    id = Column(ForeignKey("ALMA.aqua_status_history.id"), primary_key=True)
    qa2reason = Column(String(500))


class NgasFiles(Base):
    __tablename__ = "NGAS_FILES"
    __table_args__ = {"schema": "NGAS"}

    file_id = Column(String(220), primary_key=True)
    file_size = Column(Numeric(11, 0, asdecimal=True))
